import argparse
import os
import re
import sys
import subprocess

from anospp_analysis.util import *
from anospp_analysis.iplot import plot_plate_view

SUM_HAP_COLS = [
    'sample_id',
    'target',
    'reads',
    'total_reads',
    'reads_fraction',
    'nalleles',
    'seqid',
    'sample_name',
    'contamination_status',
    'contamination_confidence',
    'sseqid',
    'pident',
    'qcovs',
    'species_assignment',
    'hap_seqid',
    'plate_id',
    'well_id',
    'consensus'
]

def run_blast(plasm_hap_df, outdir, blastdb, min_pident, min_qcov):

    logging.info('running blast')

    seq_df = plasm_hap_df[['seqid', 'consensus']].drop_duplicates()

    with open(f"{outdir}/plasm_haps.fasta", "w") as output:
        for _, row in seq_df.iterrows():
            output.write(f">{row['seqid']}\n")
            output.write(f"{row['consensus']}\n")

    # Run blast and capture the output
    blast_cols = 'qseqid sseqid slen qstart qend length mismatch gapopen gaps sseq pident evalue bitscore qcovs'
    cmd = (
        f"blastn -db {blastdb} "
        f"-query {outdir}/plasm_haps.fasta "
        f"-out {outdir}/plasm_blastout.tsv "
        f"-outfmt '6 {blast_cols}' "
        f"-word_size 5 -max_target_seqs 1 -evalue 0.01"
        )
    process = subprocess.run(cmd, capture_output=True, text=True, shell=True)

    # Handle errors
    if process.returncode != 0:
        logging.error(f"An error occurred while running the blastn command: {cmd}")
        logging.error(f"Command error: {process.stderr}")
        sys.exit(1)

    blast_df = pd.read_csv(f'{outdir}/plasm_blastout.tsv', sep='\t', names=blast_cols.split())

    # not handling multiple blast hits for now
    multi_hits = blast_df.qseqid.duplicated()
    if multi_hits.any():
        multi_hits_haps = blast_df.qseqid[multi_hits].to_list()
        logging.warning(f'multiple blast hits found for {multi_hits_haps}, retaining only first hit for analysis')
        blast_df = blast_df[~ multi_hits]

    # annotate blast results
    blast_df['genus'] = blast_df.sseqid.str.split('_').str.get(0)
    blast_df['species'] = blast_df.sseqid.str.split('_').str.get(1)
    blast_df['binomial'] = blast_df['genus'] + '_' + blast_df['species']
    blast_df['species_assignment'] = blast_df['binomial']
    unknown_species = ((blast_df.pident < min_pident) | (blast_df.qcovs < min_qcov))
    blast_df.loc[unknown_species, 'species_assignment'] = 'unknown'
    blast_df['ref_seqid'] = blast_df.sseqid.str.split(':').str.get(1)

    def assign_hap_id(blast_row):
        
        # most annotations require both 100% coverage and identity    
        if blast_row.pident == 100:
            if blast_row.qcovs == 100:
                return blast_row.ref_seqid
            # M annotations require lower query coverage
            elif blast_row.ref_seqid.startswith('M') and (blast_row.qcovs >= min_qcov):
                return blast_row.ref_seqid
        
        # use per-per run seqids P1-0 -> X1-0 etc
        # seqids won't be sequential 
        hap_id_x = blast_row.qseqid.replace('P', 'X')

        return hap_id_x
        
    blast_df['hap_seqid'] = blast_df.apply(assign_hap_id, axis=1)

    return blast_df

def estimate_contamination(hap_df, comb_stats_df, min_samples, min_source_reads, max_affected_reads):
    """
    Identify potential contamination from excessive haplotype sharing between
    high coverage sample (source) and many low coverage samples (affected).

    Contamination is more likely between samples sharing plates or wells
    """

    logging.info('estimating cross-contamination')

    hap_df = hap_df[['sample_id', 'seqid', 'reads']]
    ext_hap_df = pd.merge(hap_df, comb_stats_df, on='sample_id', how='left')

    assert ~ext_hap_df['well_id'].isna().any(), 'failed to get well IDs'
    assert ~ext_hap_df['plate_id'].isna().any(), 'failed to get plate IDs'

    ext_hap_df['contamination_status'] = ''
    ext_hap_df['contamination_confidence'] = ''

    for seqid, hapid_df in ext_hap_df.groupby('seqid'):

        # status - haplotype sharing
        if (hapid_df.reads > min_source_reads).any():
            if (hapid_df.reads < max_affected_reads).sum() > min_samples:
                # source and affected data
                src_df = hapid_df.loc[hapid_df.reads > min_source_reads]
                tgt_df = hapid_df.loc[hapid_df.reads < max_affected_reads]
                # sample & hap define positions in original df
                src_haps = (ext_hap_df.sample_id.isin(src_df['sample_id']) & (ext_hap_df.seqid == seqid))
                tgt_haps = (ext_hap_df.sample_id.isin(tgt_df['sample_id']) & (ext_hap_df.seqid == seqid))
                # set contamination statuses in original df
                ext_hap_df.loc[(ext_hap_df.seqid == seqid), 'contamination_status'] = 'unclear'
                ext_hap_df.loc[src_haps, 'contamination_status'] = 'source'
                ext_hap_df.loc[tgt_haps, 'contamination_status'] = 'affected'
                # confidence - low without plate/well match
                ext_hap_df.loc[tgt_haps, 'contamination_confidence'] = 'low'
                for _, src_row in src_df.iterrows():
                    # affected samples sharing plate or well with source
                    same_plate_tgt_samples = tgt_df.loc[tgt_df.plate_id == src_row.plate_id, 'sample_id']
                    same_well_tgt_samples = tgt_df.loc[tgt_df.well_id == src_row.well_id, 'sample_id']
                    hc_tgt_samples = pd.concat([same_plate_tgt_samples, same_well_tgt_samples])
                    if len(hc_tgt_samples) > 0:
                        # sample & hap define positions in original df
                        hc_tgt_haps = (ext_hap_df.sample_id.isin(hc_tgt_samples) & (ext_hap_df.seqid == seqid))
                        # update confidence for plate/well match
                        ext_hap_df.loc[hc_tgt_haps, 'contamination_confidence'] = 'high'

    return ext_hap_df

def summarise_haplotypes(hap_df, blast_df, contam_df):

    logging.info('summarising haplotype info')

    sum_hap_df = pd.merge(hap_df, contam_df, how='left') # multiple columns to be merged
    sum_hap_df = pd.merge(sum_hap_df, blast_df, left_on='seqid', right_on='qseqid')

    sum_hap_df = sum_hap_df[SUM_HAP_COLS]

    return sum_hap_df

def summarise_samples(sum_hap_df, comb_stats_df, filters=(10,10)):

    logging.info('summarising sample info')

    sum_samples_df = comb_stats_df[[
        'sample_id',
        'sample_name',
        'lims_plate_id',
        'lims_well_id',
        'plate_id',
        'well_id'
    ]].copy().set_index('sample_id')

    sum_hap_df['reads_str'] = sum_hap_df['reads'].astype(str)
    for i, t in enumerate(PLASM_TARGETS):
        t_hap_df = sum_hap_df[sum_hap_df.target == t]
        t_sum_hap_gbs = t_hap_df.groupby('sample_id')
        sum_samples_df[f'{t}_reads_total'] = t_sum_hap_gbs['reads'].sum()
        sum_samples_df[f'{t}_reads_total'] = sum_samples_df[f'{t}_reads_total'].astype(float).fillna(0).astype(int)
        # pass criteria:
        # - read count over filter value
        # - haplotype is not high confidence affected by contamination
        t_pass_hap_gbs = t_hap_df[
            (t_hap_df.reads >= filters[i]) &
            (t_hap_df.contamination_confidence != 'high')
            ].sort_values('reads', ascending=False).groupby('sample_id')
        sum_samples_df[f'{t}_reads_pass'] = t_pass_hap_gbs['reads'].sum()
        sum_samples_df[f'{t}_reads_pass'] = sum_samples_df[f'{t}_reads_pass'].astype(float).fillna(0).astype(int)
        sum_samples_df[f'{t}_hapids_pass'] = t_pass_hap_gbs.agg({'hap_seqid': ';'.join})
        sum_samples_df[f'{t}_hapids_pass'] = sum_samples_df[f'{t}_hapids_pass'].fillna('')
        sum_samples_df[f'{t}_hapids_pass_reads'] = t_pass_hap_gbs.agg({'reads_str': ';'.join})
        sum_samples_df[f'{t}_hapids_pass_reads'] = sum_samples_df[f'{t}_hapids_pass_reads'].fillna('')
        sum_samples_df[f'{t}_species_assignments_pass'] = t_pass_hap_gbs.agg(
            {'species_assignment': ';'.join}
            )
        sum_samples_df[f'{t}_species_assignments_pass'] = sum_samples_df[f'{t}_species_assignments_pass'].fillna('')
        # contaminated haplotypes with read count over filter value
        t_contam_hap_gbs = t_hap_df[
            (t_hap_df.reads >= filters[i]) &
            (t_hap_df.contamination_status == 'affected') &
            (t_hap_df.contamination_confidence == 'high')
            ].sort_values('reads', ascending=False).groupby('sample_id')
        sum_samples_df[f'{t}_hapids_contam'] = t_contam_hap_gbs.agg({'hap_seqid': ';'.join})
        sum_samples_df[f'{t}_hapids_contam'] = sum_samples_df[f'{t}_hapids_contam'].fillna('')
        sum_samples_df[f'{t}_hapids_contam_reads'] = t_contam_hap_gbs.agg({'reads_str': ';'.join})
        sum_samples_df[f'{t}_hapids_contam_reads'] = sum_samples_df[f'{t}_hapids_contam_reads'].fillna('')
        # low coverage haplotypes
        t_locov_hap_gbs = t_hap_df[
            (t_hap_df.reads < filters[i])
            ].sort_values('reads', ascending=False).groupby('sample_id')
        sum_samples_df[f'{t}_hapids_locov'] = t_locov_hap_gbs.agg({'hap_seqid': ';'.join})
        sum_samples_df[f'{t}_hapids_locov'] = sum_samples_df[f'{t}_hapids_locov'].fillna('')
        sum_samples_df[f'{t}_hapids_locov_reads'] = t_locov_hap_gbs.agg({'reads_str': ';'.join})
        sum_samples_df[f'{t}_hapids_locov_reads'] = sum_samples_df[f'{t}_hapids_locov_reads'].fillna('')
        sum_samples_df[f'{t}_species_assignments_locov'] = t_locov_hap_gbs.agg(
            {'species_assignment': ';'.join}
            )
        sum_samples_df[f'{t}_species_assignments_locov'] = sum_samples_df[f'{t}_species_assignments_locov'].fillna('')
        

    def infer_status(sum_samples_row, targets=PLASM_TARGETS):
        # not generalised
        p1_spp_pass = set(sum_samples_row['P1_species_assignments_pass'].split(';')) - set([''])
        p2_spp_pass = set(sum_samples_row['P2_species_assignments_pass'].split(';')) - set([''])
        p1_spp_locov = set(sum_samples_row['P1_species_assignments_locov'].split(';')) - set([''])
        p2_spp_locov = set(sum_samples_row['P2_species_assignments_locov'].split(';')) - set([''])
        is_contam = (
            (len(sum_samples_row['P1_hapids_contam']) > 0) |
            (len(sum_samples_row['P2_hapids_contam']) > 0)
        )
        if len(p1_spp_pass) > 0:
            if len(p2_spp_pass) > 0:
                if p1_spp_pass == p1_spp_pass:
                    status = 'species_consistent'
                elif p1_spp_pass - p1_spp_pass == set():
                    status = 'extra_species_in_P2'
                elif p1_spp_pass - p1_spp_pass == set():
                    status = 'extra_species_in_P1'
                else:
                    status = 'species_discordant'
            else:
                # species consistent even if P2 does not pass coverage filter
                if p1_spp_pass == p2_spp_locov:
                    status = 'species_consistent_P2_locov'
                else:
                    status = 'P1_only'
        elif len(p2_spp_pass) > 0:
            if p1_spp_locov == p2_spp_pass:
                status = 'species_consistent_P1_locov'
            status = 'P2_only'
        elif is_contam:
            status = 'contamination_only'
        else:
            status = 'not_detected'

        return status

    sum_samples_df['plasmodium_detection_status'] = sum_samples_df.apply(infer_status, axis=1)

    def consensus_species(sum_samples_row, targets=PLASM_TARGETS):

        # no consensus species for statuses not considered positive
        nocall_statuses = (
            'contamination_only', 
            'P1_only', 
            'P2_only', 
            'not_detected'
            )
        if sum_samples_row['plasmodium_detection_status'] in nocall_statuses:
            return ''

        spp = set()
        for t in targets:
            tsp = set(sum_samples_row[f'{t}_species_assignments_pass'].split(';'))
            # locov only included when pass of same species is present, skip
            # tsp = tsp.union(set(sum_samples_row[f'{t}_species_assignments_locov'].split(';')))
            tsp = tsp - set([''])
            spp = spp.union(tsp)

        return ';'.join(spp)

    sum_samples_df['plasmodium_species'] = sum_samples_df.apply(consensus_species, axis=1)

    return sum_samples_df

def plasm(args):

    # Set up logging and create output directories
    setup_logging(verbose=args.verbose)

    os.makedirs(args.outdir, exist_ok=True)

    logging.info('ANOSPP plasm data import started')
    hap_df = prep_hap(args.haplotypes)
    run_id, comb_stats_df = prep_comb_stats(args.stats)

    plasm_hap_df = hap_df[hap_df['target'].isin(PLASM_TARGETS)].copy()
    
    reference_path = args.reference_path.rstrip('/')

    reference_version = reference_path.split('/')[-1]

    assert re.match(r'^plasmv\d', reference_version), f'{reference_version} not recognised as plasm ref version'

    assert os.path.isdir(reference_path), f'reference version {reference_version} does not exist at {reference_path}'


    if plasm_hap_df.shape[0] > 0:
        blastdb = f'{reference_path}/{args.blast_db_prefix}'

        blast_df = run_blast(
            plasm_hap_df, 
            args.outdir, 
            blastdb,
            args.blast_min_pident,
            args.blast_min_qcov
            )

        contam_df = estimate_contamination(
            plasm_hap_df, comb_stats_df,
            min_samples=args.contam_min_samples_affected, 
            min_source_reads=args.contam_min_reads_source, 
            max_affected_reads=args.contam_max_reads_affected
            )

        sum_hap_df = summarise_haplotypes(hap_df, blast_df, contam_df)

    # no plasmodium sequences in run - empty hap df
    else: 
        sum_hap_df = pd.DataFrame(columns=SUM_HAP_COLS)

    sum_hap_df.to_csv(f'{args.outdir}/plasm_hap_summary.tsv', sep='\t', index=False)

    sum_samples_df = summarise_samples(sum_hap_df, comb_stats_df, filters=(args.filter_p1, args.filter_p2))

    sum_samples_df['plasm_ref'] = reference_version

    out_df = sum_samples_df.drop(columns=[
        'sample_name',
        'lims_plate_id',
        'lims_well_id',
        'plate_id',
        'well_id',
        'P1_reads_total',
        'P2_reads_total'
    ]).copy()

    out_df.columns = out_df.columns.str.lower()

    out_df.to_csv(f'{args.outdir}/plasm_assignment.tsv', sep='\t')

    if args.interactive_plotting:
        for lims_plate in sum_samples_df.lims_plate_id.unique():

            logging.info(f'plotting interactive lims plate view for {lims_plate}')
            
            out_fn = f'{args.outdir}/plasm_{lims_plate}.html'
            title = f'Plasmodium species composition run {run_id}, plate {lims_plate}'
            plot_df = sum_samples_df[sum_samples_df.lims_plate_id == lims_plate]
            plot_plate_view(plot_df, out_fn, reference_path, title)

    logging.info('ANOSPP plasm complete')

    return


def main():

    parser = argparse.ArgumentParser("Plasmodium ID assignment for ANOSPP data")
    parser.add_argument('-a', '--haplotypes', help='Haplotypes tsv file generated by prep', required=True)
    parser.add_argument('-s', '--stats', help='Comb stats tsv file generated by prep', required=True)
    parser.add_argument('-r', '--reference_path', 
                        help='Path to plasm reference directory, expected to end with e.g. plasmv1', 
                        required=True)
    parser.add_argument('-o', '--outdir', help='Output directory. Default: qc', default='plasm')
    parser.add_argument('--blast_db_prefix', 
                        help='Blast database prefix in reference index', 
                        default='plasmomito_P1P2_DB_v1.0')
    parser.add_argument('--blast_min_pident', 
                        help=('Minimum blast percent identity for haplotype assignment to species. '
                              'Default: 99 - corresponds to 2 SNPs or indels per ~210bp target'),
                        default=99, type=float)
    parser.add_argument('--blast_min_qcov', 
                        help='Minimum blast alignment query coverage for haplotype assignment to species. Default: 96',  
                        default=96, type=int)
    parser.add_argument('--contam_min_reads_source', 
                        help='Minimum number of reads in a source sample for same plate/well contamination. Default: 10000',  
                        default=10000, type=int)
    parser.add_argument('--contam_max_reads_affected', 
                        help='Maximum number of reads in a sample affected by same plate/well contamination. Default: 10000',  
                        default=100, type=int)
    parser.add_argument('--contam_min_samples_affected', 
                        help='Minimum number of samples affected by plate/well contamination to be considered. Default: 4',  
                        default=4, type=int)
    parser.add_argument('-f', '--filter_p1',
                        help='Minimum read support for P1 haplotypes to be included in sample summary. Default: 10',
                        default=10, type=int)
    parser.add_argument('-g', '--filter_p2', 
                        help='Minimum read support for P2 haplotypes to be included in sample summary. Default: 10', 
                        default=10, type=int)
    parser.add_argument('-i', '--interactive_plotting', 
                        help='Create interactive plots of species composition across plates', 
                        action='store_true', default=False)
    parser.add_argument('-v', '--verbose', 
                        help='Include INFO level log messages', action='store_true')


    args = parser.parse_args()
    args.outdir=args.outdir.rstrip('/')

    plasm(args)


if __name__ == '__main__':
    main()




###imports for bokeh
from bokeh.plotting import output_file, save
from bokeh.plotting import figure, output_file
from bokeh.models import ColumnDataSource, Span, Range1d
from bokeh.transform import factor_cmap
from bokeh.layouts import gridplot
from bokeh.models.glyphs import Text, Rect
from bokeh.models.tools import HoverTool
from bokeh.transform import dodge

import os
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

from anospp_analysis.util import *


def view_alignment(aln, aln_view_fn, fontsize='9pt', plot_width=1200):
    '''Bokeh sequence alignment view'''

    def get_colors(seqs):
        '''make colors for bases in sequence'''
        text = [i for s in list(seqs) for i in s]
        clrs =  {'a':'red','t':'green','g':'orange','c':'blue','-':'white', 'n':'black'}
        colors = [clrs[i] for i in text]
        return colors

    output_file(filename=aln_view_fn, title='Static HTML file')

    #make sequence and id lists from the aln object
    seqs = [rec.seq for rec in (aln)]
    ids = [rec.id for rec in aln]    
    text = [i for s in list(seqs) for i in s]
    colors = get_colors(seqs)    
    N = len(seqs[0])
    S = len(seqs)    
    width = .4

    x = np.arange(1,N+1)
    y = np.arange(0,S,1)
    #creates a 2D grid of coords from the 1D arrays
    xx, yy = np.meshgrid(x, y)
    #flattens the arrays
    gx = xx.ravel()
    gy = yy.flatten()
    #use recty for rect coords with an offset
    recty = gy+.5
    h= 1/S
    #now we can create the ColumnDataSource with all the arrays
    source = ColumnDataSource(dict(x=gx, y=gy, recty=recty, text=text, colors=colors))
    plot_height = len(seqs)*15+50
    x_range = Range1d(0,N+1, bounds='auto')
    if N>100:
        viewlen=100
    else:
        viewlen=N
    #view_range is for the close up view
    view_range = (0,viewlen)
    tools='xpan, xwheel_zoom, reset, save'

    #entire sequence view (no text, with zoom)
    p = figure(title=None, width = plot_width, height=50,
               x_range=x_range, y_range=(0,S), tools=tools,
               min_border=0, toolbar_location='below')
    rects = Rect(x='x', y='recty',  width=1, height=1, fill_color='colors',
                 line_color=None, fill_alpha=0.6)
    p.add_glyph(source, rects)
    p.yaxis.visible = False
    p.grid.visible = False  

    #sequence text view with ability to scroll along x axis
    p1 = figure(title=None, width=plot_width, height=plot_height,
                x_range=view_range, y_range=ids, tools='xpan,reset',
                min_border=0, toolbar_location='below')#, lod_factor=1)          
    glyph = Text(x='x', y='y', text='text', text_align='center',text_color='black',
                text_font='monospace',text_font_size=fontsize)
    rects = Rect(x='x', y='recty',  width=1, height=1, fill_color='colors',
                line_color=None, fill_alpha=0.4)
    p1.add_glyph(source, glyph)
    p1.add_glyph(source, rects)

    p1.grid.visible = False
    p1.xaxis.major_label_text_font_style = 'bold'
    p1.yaxis.minor_tick_line_width = 0
    p1.yaxis.major_tick_line_width = 0

    p = gridplot([[p],[p1]], toolbar_location='below')
    save(p)

def plot_plate_view(df, out_fn, reference_path, title=None):

    '''
    Plots a plate map for a given plate and Plasmodium type.

    Args:
    - df (pandas.DataFrame): DataFrame to plot.
    - out_fn (str): name of the file to save the plot
    - reference_path (str): path to the reference directory
    - title (str): title for the plot. Default is None.

    Returns:
    None.
    '''

    assert os.path.isdir(reference_path), f'{reference_path} reference directory does not exist'

    # set the output filename
    output_file(out_fn)

    #extract the column and generate the row values
    cols = list('ABCDEFGHIJKLMNOP')
    rows = [str(x) for x in range(1, 25)]
    df = df.reset_index(drop=False)
    df['col'] = df['lims_well_id'].str[0]
    df['row'] = df['lims_well_id'].str[1:]
    
    # display values
    df['P1_hapids_disp'] = df['P1_hapids_pass'].str.replace(',.*', '...', regex=True)
    df['P2_hapids_disp'] = df['P2_hapids_pass'].str.replace(',.*', '...', regex=True)
    df['comb_hapids_disp'] = df['P1_hapids_disp'] + '\n' + df['P2_hapids_disp']

    #load the dataframe into the source
    source = ColumnDataSource(df)

    #set up the figure
    p = figure(
        width=1300,
        height=600,
        title=title,
        x_range=rows,
        y_range=list(reversed(cols)),
        toolbar_location=None,
        tools=[HoverTool(), 'pan', 'wheel_zoom', 'reset']
        )

    # add grid lines
    for v in range(len(rows)):
        line_width = 2 if (v % 2 == 0) else 1
        vline = Span(location=v, dimension='height', line_color='black', line_width=line_width)
        p.renderers.extend([vline])

    for h in range(len(cols)):
        line_width = 2 if (h % 2 == 0) else 1
        hline = Span(location=h, dimension='width', line_color='black', line_width=line_width)
        p.renderers.extend([hline])

    #load colors
    if not os.path.isfile(f'{reference_path}/species_colours.csv'):
        logging.warning('no colors defined for plotting')
        cmap = {}
    else:
        colors = pd.read_csv(f'{reference_path}/species_colours.csv')
        cmap = dict(zip(colors['species'], colors['color']))

    
    for index, row in df.iterrows():
        # Assign grey color to data with more than one species
        if len(row['plasmodium_species'].split(',')) > 1:
            cmap[row['plasmodium_species']] = '#cfcfcf'
        # Assign white color to data with no species
        elif len(row['plasmodium_species']) == 0:
            cmap[row['plasmodium_species']] = '#ffffff'

    #add the rectangles
    p.rect(
        'row',
        'col',
        0.95,
        0.95,
        source=source,
        fill_alpha=.9,
        legend_field='plasmodium_species',
        color=factor_cmap('plasmodium_species', palette=list(cmap.values()), factors=list(cmap.keys()))
        )

    #add the species count text for each field
    text_props = {'source': source, 'text_align': 'left', 'text_baseline': 'middle'}
    x = dodge('row', -0.4, range=p.x_range)
    r = p.text(x=x, y='col', text='comb_hapids_disp', **text_props)
    r.glyph.text_font_size = '10px'
    r.glyph.text_font_style = 'bold'

    #set up the hover value
    p.add_tools(HoverTool(tooltips=[
        ('sample id', '@{sample_id}'),
        ('Parasite species', '@plasmodium_species'),
        ('Detection status', '@plasmodium_detection_status'),
        ('P1 total reads', '@P1_reads_total'),
        ('P1 QC pass reads', '@P1_reads_pass'),
        ('P1 QC pass haplotype IDs', '@P1_hapids_pass'),
        ('P1 reads per QC pass haplotype', '@P1_hapids_pass_reads'),
        ('P1 species assignments for pass haplotypes', '@P1_species_assignments_pass'),
        ('P1 contamination haplotype IDs', '@P1_hapids_contam'),
        ('P1 reads per contamination haplotype', '@P1_hapids_contam_reads'),
        ('P1 low coverage haplotype IDs', '@P1_hapids_locov'),
        ('P1 reads per low coverage haplotype', '@P1_hapids_locov_reads'),
        ('P2 total reads', '@P2_reads_total'),
        ('P1 QC pass reads', '@P2_reads_pass'),
        ('P2 QC pass haplotype IDs', '@P2_hapids_pass'),
        ('P2 reads per QC pass haplotype', '@P2_hapids_pass_reads'),
        ('P2 species assignments for pass haplotypes', '@P2_species_assignments_pass'),
        ('P2 contamination haplotype IDs', '@P2_hapids_contam'),
        ('P2 reads per contamination haplotype', '@P2_hapids_contam_reads'),
        ('P2 low coverage haplotype IDs', '@P2_hapids_locov'),
        ('P2 reads per low coverage haplotype', '@P2_hapids_locov_reads'),
    ]))

    #set up the rest of the figure and save the plot
    p.outline_line_color = 'black'
    p.grid.grid_line_color = None
    p.axis.axis_line_color = 'black'
    p.axis.major_tick_line_color = None
    p.axis.major_label_standoff = 0
    p.legend.orientation = 'vertical'
    p.legend.click_policy='hide'
    p.add_layout(p.legend[0], 'right') 
    save(p)


import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from collections import OrderedDict
from scipy.optimize import curve_fit
import math
import os
import argparse

from anospp_analysis.util import *

def plot_target_balance(hap_df, run_id):

    logging.info('plotting targets balance')

    reads_per_sample_target = hap_df \
        .groupby(['sample_id', 'target'], observed=False)['reads'].sum().reset_index()
    
    figsize = (hap_df['target'].nunique() * 0.3, 6)
    fig, ax = plt.subplots(1, 1, figsize=figsize, constrained_layout=True)
    fig.suptitle(f'Amplicon coverage distribution per sample for run {run_id}')
    sns.stripplot(
        data=reads_per_sample_target,
        x = 'target',
        y = 'reads',
        hue = 'target',
        alpha = .1,
        jitter = .3,
        ax = ax
        )
    # ax.get_legend().remove()
    ax.set_yscale('log')
    # ax.set_ylim(bottom=0.5)
    ax.set_ylabel('reads')
    ax.set_xlabel('target')
    ax.axhline(10, c='silver', alpha=.5)
    ax.tick_params(axis='x', rotation=90)

    return fig, ax

def plot_allele_balance(hap_df, run_id, anospp=True):
    
    logging.info('plotting allele balance and coverage')

    if anospp:
        hap_df = hap_df[hap_df.target.isin(MOSQ_TARGETS)]

    is_het = (hap_df.reads_fraction < 1)
    het_frac = hap_df[is_het].reads_fraction
    het_reads_log = hap_df[is_het].reads.apply(lambda x: np.log10(x))
    het_plot = sns.jointplot(
        x=het_frac,
        y=het_reads_log,
        kind="hist",
        height=8
        )
    het_plot.ax_joint.set_ylabel('reads (log10)')
    het_plot.ax_joint.set_xlabel('allele fraction')
    het_plot.ax_joint.axhline(1, c='silver', alpha=.5)
    het_plot.ax_joint.axvline(0.1, c='silver', alpha=.5)
    # funky title setting
    extra = ' in mosquito targets' if anospp else ''
    het_plot.fig.suptitle(f'Allele fractions and coverage{extra} for run {run_id}')
    het_plot.fig.tight_layout()
    het_plot.fig.subplots_adjust(top=0.95) # Reduce plot to make room
    
    return het_plot

def plot_well_balance(comb_stats_df, run_id):

    logging.info('plotting wells balance')

    fig, ax = plt.subplots(figsize=(14, 4))
    fig.suptitle(f'Reads per well for run {run_id}')
    sns.boxplot(
            data=comb_stats_df,
            x = 'well_id',
            y = 'total_reads',
            ax = ax
            )
    ax.set_yscale('log')
    ax.tick_params(axis='x', rotation=90)

    return fig, ax

def plot_sample_filtering(comb_stats_df, run_id, anospp=True):
    
    logging.info('plotting per-sample filtering barplots')

    # comb_stats_df colname : legend label
    dada2_cols = OrderedDict([
        ('total_reads', 'removed as readthrough'),
        ('dada2_input_reads', 'removed by filterAndTrim'), 
        ('dada2_filtered_reads', 'removed by denoising'),
        ('dada2_denoised_reads', 'removed by merging'), 
        ('dada2_merged_reads', 'removde by rmchimera'), 
        ('dada2_nonchim_reads', 'unassigned to amplicons'),
        # legacy post-filter disabled in prod
        # ('dada2_final_reads', 'unassigned to amplicons'),
        ])

    if anospp:
        dada2_cols['target_reads'] = 'Plasmodium reads'
        dada2_cols['raw_mosq_reads'] = 'mosquito reads'
    else:
        dada2_cols['target_reads'] = 'target reads'
    
    plates = comb_stats_df.plate_id.unique()
    nplates = len(plates)
    fig, axs = plt.subplots(nplates, 1, figsize=(20, 4 * nplates), constrained_layout=True)
    fig.suptitle(f'Read filtering stats for run {run_id}', fontsize=20)
    if nplates == 1:
        axs = [axs]
    for plate, ax in zip(plates, axs):
        plot_df = comb_stats_df[comb_stats_df.plate_id == plate].copy()
        plot_df['well_id'] = well_ordering(plot_df['well_id'])
        plot_df.sort_values(by='well_id', inplace=True)
        for i, col in enumerate(dada2_cols.keys()):
            sns.barplot(
                x='sample_id',
                y=col,
                data=plot_df,
                color=sns.color_palette()[i],
                ax=ax,
                label=dada2_cols[col],
                width=1
                )
        ax.set_xticks(range(plot_df.shape[0]))
        ax.set_xticklabels(plot_df['sample_name'])
        ax.set_xlabel('sample_name')
        ax.tick_params(axis='x', rotation=90)
        ax.set_yscale('log')
        ax.set_ylim(bottom=0.5, top=max(comb_stats_df['total_reads']))
        ax.set_ylabel('reads')
        ax.legend(loc='upper left', bbox_to_anchor=(1,1), fontsize=14)
        ax.set_title(plate)

    return fig, axs

def plot_plate_stats(comb_stats_df, run_id, lims_plate=False):

    plate_col = 'lims_plate_id' if lims_plate else 'plate_id'
    size = '384-well' if lims_plate else '96-well'

    logging.info('plotting plate stats')
    
    fig, axs = plt.subplots(
        3, 1,
        figsize=(10,15),
        constrained_layout=True
        )
    fig.suptitle(f'Key {size} plate performance stats for run {run_id}')

    sns.stripplot(
        data=comb_stats_df,
        y='target_reads',
        x=plate_col,
        hue=plate_col,
        alpha=.3,
        jitter=.35,
        ax=axs[0]
        )
    axs[0].set_yscale('log')
    axs[0].set_ylim(bottom=1)
    axs[0].set_xlabel('')
    axs[0].set_xticklabels([])
    # 1000 reads cutoff
    axs[0].axhline(1000, c='silver', alpha=.5)
    
    sns.stripplot(
        data=comb_stats_df,
        y='raw_mosq_targets_recovered',
        x=plate_col,
        hue=plate_col,
        alpha=.3,
        jitter=.35,
        ax=axs[1]
        )
    axs[1].set_ylim(bottom=0, top=62)
    axs[1].set_xticklabels([])
    axs[1].set_xlabel('')
    # 10/50 targets cutoff - nn/vae
    axs[1].axhline(10, c='silver', alpha=.5)
    axs[1].axhline(50, c='silver', alpha=.5)

    sns.stripplot(
        data=comb_stats_df,
        y='overall_filter_rate',
        x=plate_col,
        hue=plate_col,
        alpha=.3,
        jitter=.35,
        ax=axs[2]
        )
    axs[2].set_ylim(bottom=0, top=1)
    # 50% filtering cutoff
    axs[2].axhline(.5, c='silver', alpha=.5)
    # for ax in axs:
    #     ax.get_legend().remove()
    plt.xticks(rotation=90)

    return fig, axs

def plot_plate_summaries(comb_stats_df, run_id, lims_plate=False):

    plate_col = 'lims_plate_id' if lims_plate else 'plate_id'
    size = '384-well' if lims_plate else '96-well'

    logging.info(f'plotting success summaries by {plate_col}')
    
    # success rate definition
    comb_stats_df['over 1000 mosquito reads'] = comb_stats_df.raw_mosq_reads > 1000
    comb_stats_df['over 10 targets'] = comb_stats_df.targets_recovered > 10
    comb_stats_df['over 50% reads retained'] = comb_stats_df.overall_filter_rate > .5

    plates = comb_stats_df[plate_col].unique()
    nplates = comb_stats_df[plate_col].nunique()

    sum_df = comb_stats_df.groupby(plate_col) \
        [['over 1000 mosquito reads', 'over 10 targets', 'over 50% reads retained']].sum()
    y = comb_stats_df.groupby(plate_col)['over 1000 mosquito reads'].count()
    sum_df = sum_df.divide(y, axis=0).reindex(plates)

    fig, ax = plt.subplots(
        1, 1, 
        figsize=(nplates * .5 + 2.5, 4),
        constrained_layout=True
        )
    fig.suptitle(f'{size} plate success rate for run {run_id}', fontsize=10)
    sns.heatmap(sum_df.T, annot=True, ax=ax, vmax=1, vmin=0)
    plt.xticks(rotation=90)

    return fig, ax

def plot_sample_success(comb_stats_df, run_id, anospp=True):

    logging.info('plotting sample success')

    ycol = 'raw_mosq_reads' if anospp else 'target_reads'
    tgtcol = 'raw_mosq_targets_recovered' if anospp else 'targets_recovered'

    fig, axs = plt.subplots(1, 2, figsize=(12,6), constrained_layout=True)
    fig.suptitle(f'Key stats covariation for run {run_id}')
    for xcol, ax in zip((tgtcol, 'overall_filter_rate'), axs):
        sns.scatterplot(
            data=comb_stats_df,
            x=xcol,
            y=ycol,
            hue='plate_id',
            alpha=.5, 
            ax=ax
            )
        ax.set_yscale('log')
        ax.set_ylim(bottom=1)
        ax.axhline(1000, c='silver', alpha=.5)
    
    axs[0].set_xlim(left=0)
    if anospp:
        axs[0].axvline(10, c='silver', alpha=.5)
        axs[0].axvline(50, c='silver', alpha=.5)
        axs[0].set_xlim(right=62)
    axs[1].axvline(.5, c='silver', alpha=.5)
    axs[1].set_xlim(left=0, right=1)
    axs[1].get_legend().remove()

    return fig, axs

def plot_plasm_balance(comb_stats_df, run_id):

    logging.info('plotting Plasmodium read balance')

    max_plasm_reads = max(
        max(
            comb_stats_df.p1_reads.max(), 
            comb_stats_df.p2_reads.max()
            ),
        1)

    fig, ax = plt.subplots(1, 1, figsize=(5, 5), constrained_layout=True)
    fig.suptitle(f'P1/P2 coverage balance for run {run_id}')
    sns.scatterplot(
        # display P1 or P2-only samples
        data=comb_stats_df.replace(0, 0.5),
        x='p1_reads',
        y='p2_reads',
        hue='plate_id',
        alpha=.5, 
        ax=ax
        )
    # hard filter cutoff
    ax.axhline(10, c='silver', alpha=.5)
    ax.axvline(10, c='silver', alpha=.5)
    # contamination affected sample cutoff
    ax.axhline(100, c='green', alpha=.5, linestyle='dashed')
    ax.axvline(100, c='green', alpha=.5, linestyle='dashed')
    # contamination source sample cutoff
    ax.axhline(10000, c='red', alpha=.5, linestyle='dashed')
    ax.axvline(10000, c='red', alpha=.5, linestyle='dashed')
    ax.plot(
        [0.4, max_plasm_reads],
        [0.4, max_plasm_reads],
        color='silver',
        linestyle='dashed',
        alpha=.5
        )
    ax.set_yscale('log')
    ax.set_xscale('log')

    return fig, ax

def plot_plate_heatmap(comb_stats_df, col, run_id, lims_plate=False, **heatmap_kwargs):

    plate_col = 'lims_plate_id' if lims_plate else 'plate_id'
    well_col = 'lims_well_id' if lims_plate else 'well_id'
    size = '384-well' if lims_plate else '96-well'
    plate_width = 14 if lims_plate else 7
    plate_height = 8 if lims_plate else 5
    ncols = 1 if lims_plate else 2

    logging.info(f'plotting heatmap for {col} by {plate_col}')

    plates = comb_stats_df[plate_col].unique()
    nplates = comb_stats_df[plate_col].nunique()
    # ceil
    nrows = -(nplates // -ncols)

    comb_stats_df['row'] = comb_stats_df[well_col].str.slice(0, 1)
    comb_stats_df['col'] = comb_stats_df[well_col].str.slice(1).astype(int)

    fig, axs = plt.subplots(
        nrows, ncols,
        figsize=(plate_width * ncols, plate_height * nrows), 
        constrained_layout=True
        )
    fig.suptitle(f'Heatmap of {col} for {size} plates in run {run_id}', fontsize=16)
    if nplates == 1:
        axs = np.array([axs])
    for plate, ax in zip(plates, axs.flatten()):
        pdf = comb_stats_df[comb_stats_df[plate_col] == plate]
        hdf = pdf.pivot(index='row', columns='col', values=col)
        # read counts adjustments
        if 'fmt' in heatmap_kwargs.keys():
            if heatmap_kwargs['fmt'] == '':
                # human formatted labels
                heatmap_kwargs['annot'] = hdf.map(human_format)
                # handling of zero counts
                hdf = hdf.replace(0, 0.1)
        sns.heatmap(hdf, ax=ax, **heatmap_kwargs)
        if lims_plate:
            ax.hlines([i * 2 for i in range(9)], 0, 24, colors='k')
            ax.vlines([j * 2 for j in range(13)], 0, 16, colors='k')
        title = f'{plate} {col}'
        ax.set_title(title)

    return fig, axs

def plot_het_cov(hap_df, title='Total', run_id=None):

    logging.info('plotting heterozygosity vs coverage')

    nalleles_df = hap_df.pivot_table(
        index='sample_id',
        columns='target',
        values='nalleles',
        aggfunc='max',
        fill_value=0,
        observed=False)
    het_df = pd.DataFrame({
        'reads':hap_df.groupby(['sample_id'])['reads'].sum(),
        'targets':hap_df.groupby(['sample_id'])['target'].nunique(),
        'het_targets':(nalleles_df > 1).sum(axis=1)
    })
    het_df['het_rate'] = het_df['het_targets'] / het_df['targets']

    def logistic(x, L, k, x0):
        """
        Logistic function that can plateau:
        L = maximum value (plateau)
        k = growth rate
        x0 = the x-value of the sigmoid's midpoint
        """
        return L / (1 + np.exp(-k * (x - x0)))
    # fit the logistic curve to the data using curve_fit
    popt, _ = curve_fit(logistic, het_df.reads.fillna(-1), het_df.het_rate, p0=[.4, 0.0004, 3000])
    # popt contains the optimal parameters L, k, x0
    L, k, x0 = popt
    # set graph limit to nearest 1000
    xmax=math.ceil(float(het_df.reads.max()) / 1000) * 1000
    ymax=np.max(het_df.het_rate)
    x = np.arange(xmax) * 100
    y = logistic(x, L, k, x0)
    # cutoff values
    cutoffs = {}
    # for r in (.90,.95,.99):
    # find the index of 0.9  of limit of curve-fitted line and use the same index to find x value
    y_index=len(list(filter(lambda x: float(x) < float(L) * 0.9, list(y))))
    # transform into read counts
    x_threshold = x[y_index]

    fig, ax = plt.subplots(1, 1, figsize=(6,4))

    ax.scatter(het_df.reads, het_df.het_rate, alpha=.1)
    ax.plot(x, y, c='r', label=f'Logistic het, max {L:.3f}')
    ax.vlines(x_threshold, 0, 0.95 * ymax, color='k', ls=':', label=f'Raw reads at 90% het: {x_threshold}')
    ax.set_xlim(-500,xmax)
    # ax.set_xscale('log')
    ax.legend()
    ax.set_ylabel('heterozygosity rate')
    ax.set_xlabel('reads')
    title = f'{title} reads and heterozygosity'
    if run_id is not None:
        title += f' for run {run_id}'
    ax.set_title(title)

    return fig, ax, L, x_threshold

def plot_cov(comb_stats_df, run_id):

    logging.info('plotting coverage')

    fig, axs = plt.subplots(2, 1, figsize=(10, 10), sharex=True)

    for col, ax in zip(['total_reads', 'raw_mosq_reads'], axs):

        data = comb_stats_df[col]
        
        # Calculate statistics
        mean_val = data.mean()
        median_val = data.median()
        quantiles = data.quantile([0.25, 0.75, 0.95])
        q25, q75, q95 = quantiles[0.25], quantiles[0.75], quantiles[0.95]
        
        # Alt plotting options
        # ax.set_xscale('log')
        # sns.kdeplot(x=data.replace(0,0.5), ax=ax)
        # sns.boxenplot(x=data.replace(0,0.5), ax=ax, alpha=.1)
        # sns.stripplot(x=data.replace(0,0.5), jitter=False, ax=ax)

        # Plot histogram
        ax.hist(data, bins=20, color='skyblue', edgecolor='black', alpha=0.7)
        
        # Add horizontal lines for mean, median, and quantiles
        ax.axvline(mean_val, color='red', linestyle='--', linewidth=1.5, label=f'Mean: {mean_val:.0f}')
        ax.axvline(median_val, color='green', linestyle='-', linewidth=1.5, label=f'Median: {median_val:.0f}')
        ax.axvline(q25, color='purple', linestyle='-.', linewidth=1.5, label=f'25% Quantile: {q25:.0f}')
        ax.axvline(q75, color='orange', linestyle='-.', linewidth=1.5, label=f'75% Quantile: {q75:.0f}')
        ax.axvline(q95, color='blue', linestyle='-.', linewidth=1.5, label=f'95% Quantile: {q95:.0f}')
        
        # Add labels and legend
        ax.set_title(f'Run {run_id} {col}')
        ax.set_xlabel(col)
        ax.set_ylabel('Frequency')
        ax.legend()

    return fig, axs

def qc(args):

    setup_logging(verbose=args.verbose)

    os.makedirs(args.outdir, exist_ok=True)
    
    logging.info('ANOSPP QC data import')

    hap_df = prep_hap(args.haplotypes)
    
    run_id, comb_stats_df = prep_comb_stats(args.stats)
    
    logging.info('plotting QC')
    
    if hap_df['target'].isin(CUTADAPT_TARGETS).all():
        anospp = True
        logging.info('only ANOSPP targets detected, plotting all ANOSPP QC plots')
    else:
        anospp = False
        logging.warning('non-ANOSPP targets detected, plotting only generic QC plots')

    fig, _ = plot_cov(comb_stats_df, run_id)
    fig.savefig(f'{args.outdir}/coverage.png')

    fig, _ = plot_target_balance(hap_df, run_id)
    fig.savefig(f'{args.outdir}/target_balance.png')

    # allele balance plotted for mosuqito targets only
    fig = plot_allele_balance(hap_df, run_id, anospp=anospp)
    fig.savefig(f'{args.outdir}/allele_balance.png')

    fig, _ = plot_well_balance(comb_stats_df, run_id)
    fig.savefig(f'{args.outdir}/well_balance.png')

    fig, _ = plot_sample_filtering(comb_stats_df, run_id, anospp=anospp)
    fig.savefig(f'{args.outdir}/filter_per_sample.png')

    # set of plots tweaked for anospp only 
    if anospp:
        # disabled as logistic fit does not work prior to filtering, 
        # post-filtering plot is generated as part of NN script
        # mosq_hap_df = hap_df[hap_df['target'].isin(MOSQ_TARGETS)]
        # fig, _, _, _ = plot_het_cov(mosq_hap_df, title='Raw mosquito', run_id=run_id)
        # fig.savefig(f'{args.outdir}/het_cov.png')

        fig, _ = plot_plate_stats(comb_stats_df, run_id, lims_plate=False)
        fig.savefig(f'{args.outdir}/plate_stats.png')

        fig, _ = plot_plate_summaries(comb_stats_df, run_id, lims_plate=False)
        fig.savefig(f'{args.outdir}/plate_summaries.png')

        fig, _ = plot_plate_summaries(comb_stats_df, run_id, lims_plate=True)
        fig.savefig(f'{args.outdir}/lims_plate_summaries.png')

        fig, _ = plot_sample_success(comb_stats_df, run_id, anospp=anospp)
        fig.savefig(f'{args.outdir}/sample_success.png')

        fig, _ = plot_plasm_balance(comb_stats_df, run_id)
        fig.savefig(f'{args.outdir}/plasm_balance.png')
    else:
        fig, _, _, _ = plot_het_cov(hap_df, title='Total', run_id=run_id)
        fig.savefig(f'{args.outdir}/het_cov.png')

    if anospp:
        heatmap_cols = [
            'p1_reads',
            'p2_reads',
            'total_reads', 
            'target_reads',
            'overall_filter_rate',
            'raw_mosq_targets_recovered',
            'raw_multiallelic_mosq_targets',
            'unassigned_asvs'
            ]
    else:
        heatmap_cols = [
            'total_reads', 
            'target_reads',
            'overall_filter_rate',
            'targets_recovered',
            'unassigned_asvs'
            ]

    for col in heatmap_cols:
        for lims_plate in (True, False):
            # re-init heatmap args for each plot
            heatmap_kwargs = {
                    'annot':True,
                    'cmap':'coolwarm'
                }
            if col == 'raw_mosq_targets_recovered':
                heatmap_kwargs['vmin'] = 0
                heatmap_kwargs['vmax'] = 62
            elif col == 'raw_multiallelic_mosq_targets':
                heatmap_kwargs['vmin'] = 0
                heatmap_kwargs['vmax'] = max(comb_stats_df[col])
            elif col == 'unassigned_asvs':
                heatmap_kwargs['vmin'] = 0
                heatmap_kwargs['vmax'] = max(comb_stats_df[col])
            elif col == 'overall_filter_rate':
                heatmap_kwargs['vmin'] = 0
                heatmap_kwargs['vmax'] = 1
                heatmap_kwargs['fmt'] = '.2f'
            # read counts expected in all other cases
            else:
                # log-transform colour axis
                # vmin vmax set here
                heatmap_kwargs['norm'] = LogNorm(
                    vmin=0.1,
                    vmax=max(max(comb_stats_df[col]), 0.1))
                # auto-apply human_format to annot
                heatmap_kwargs['fmt'] = '' 
            

            fig, _ = plot_plate_heatmap(
                comb_stats_df,
                col=col,
                run_id=run_id,
                lims_plate=lims_plate,
                **heatmap_kwargs
                )
            if lims_plate:
                plate_hm_fn  = f'{args.outdir}/lims_plate_hm_{col}.png' 
            else:
                plate_hm_fn  = f'{args.outdir}/plate_hm_{col}.png' 
            fig.savefig(plate_hm_fn)
            plt.close(fig)

    logging.info('ANOSPP QC complete')


def main():
    
    parser = argparse.ArgumentParser("QC for ANOSPP sequencing data")
    parser.add_argument('-a', '--haplotypes', 
                        help='Haplotypes tsv file generated by prep script', 
                        required=True)
    parser.add_argument('-s', '--stats', 
                        help='Stats tsv file generated by prep script',
                        required=True)
    parser.add_argument('-o', '--outdir', help='Output directory. Default: qc', default='qc')
    parser.add_argument('-v', '--verbose', 
                        help='Include INFO level log messages', action='store_true')

    args = parser.parse_args()
    args.outdir=args.outdir.rstrip('/')

    qc(args)


if __name__ == '__main__':
    main()


import vae
import nn
import pandas as pd
import numpy as np
from scipy.spatial import ConvexHull

def test_select_samples():
    assignment_df = pd.read_csv('test_data/output/nn_assignment.tsv', sep='\t')
    hap_df = pd.read_csv('test_data/output/non_error_haplotypes.tsv', sep='\t')

    result = vae.select_samples(
        assignment_df,
        hap_df,
        'int'	,
        'Anopheles_gambiae_complex',
        50
    )

    assert len(result[0]) == 723
    assert (result[1]).shape == (63138, 9)

def test_prep_sample_kmer_table():
    hap_df = pd.read_csv('test_data/output/non_error_haplotypes.tsv', sep='\t')
    kmers_unique_seqs = nn.construct_unique_kmer_table(hap_df, 8)
    parsed_seqids = nn.parse_seqids_series(hap_df.loc[hap_df.sample_id=='DN806197N_A1',\
                                                      'seqid'])

    result = vae.prep_sample_kmer_table(
        kmers_unique_seqs, 
        parsed_seqids
    )

    assert result.shape == (65536,)
    assert result.sum() == 18374

def test_prep_kmers():
    hap_df = pd.read_csv('test_data/output/non_error_haplotypes.tsv', sep='\t')
    vae_samples = np.array(['DN806197N_A1', 'DN806197N_A2', 'DN806197N_A3', 'DN806197N_A4', \
                            'DN806197N_A5', 'DN806197N_A6', 'DN806197N_A7', 'DN806197N_A8', \
                            'DN806197N_A9', 'DN806197N_A10', 'DN806197N_A11', 'DN806197N_A12'])
    vae_hap_df = hap_df.query('sample_id in @vae_samples')

    result = vae.prep_kmers(
        vae_hap_df,
        vae_samples,
        8
    )

    assert result.shape == (12, 65536)
    assert (result.sum(axis=1) == np.array([18374, 18628, 18357, 18634, 18364, 18326, 18613, \
                                            18627, 18629, 18619, 18655, 18354])).all()

def test_predict_latent_pos():
    hap_df = pd.read_csv('test_data/output/non_error_haplotypes.tsv', sep='\t')
    vae_samples = np.array(['DN806197N_A1', 'DN806197N_A10', 'DN806197N_A11', 'DN806197N_A12', \
                            'DN806197N_A2', 'DN806197N_A3', 'DN806197N_A4', 'DN806197N_A5', \
                            'DN806197N_A6', 'DN806197N_A7', 'DN806197N_A8', 'DN806197N_A9'])
    vae_hap_df = hap_df.query('sample_id in @vae_samples')
    kmer_table = vae.prep_kmers(vae_hap_df, vae_samples, 8)
    comparison = pd.read_csv("test_data/comparisons/latent_coordinates.tsv", sep='\t', \
                             index_col=0)

    result = vae.predict_latent_pos(
        kmer_table, 
        vae_samples, 
        8,
        'ref_databases/gcrefv1/_weights.hdf5'
    )
    assert (result.index.values == comparison.index.values).all()
    assert (np.abs(result.mean1.values - comparison.mean1.values) < 0.001).all()
    assert (np.abs(result.mean2.values - comparison.mean2.values) < 0.001).all()
    assert (np.abs(result.mean3.values - comparison.mean3.values) < 0.001).all()

def test_find_normal_edge_simplex():
    a,b,c = np.array([1,0,0]), np.array([4,0,0]), np.array([3,4,0])

    result = vae.find_normal_edge_simplex(
        a,
        b,
        c
    )

    assert (result[0] == np.array([0,1,0])).all()
    assert (result[1] == np.array([1,0,0])).all()

def test_check_half_space():

    n, o = np.array([0,1,0]), np.array([1,0,0])
    points = np.array([[2,-1,11], [4,2,2], [-33,5,-1], [2,0,0]])

    result = np.array([
        vae.check_half_space(p,n,o) for p in points
    ])

    expected_result = np.array([
        [False, True, True, True]
    ])

    assert (result == expected_result).all()
  
def test_compute_distance_to_vertex():

    p = np.array([3,1,2])
    v = np.array([4,0,0])

    result = vae.compute_distance_to_vertex(
        p, 
        v
    )

    assert result == np.sqrt(6)

def test_compute_distance_to_edge():
    v, w = np.array([1,0,0]), np.array([4,0,0])
    points = np.array([[1,2,0], [3,-3,-3], [0,-1,0], [6,1,-1]])

    result = np.array([
        vae.compute_distance_to_edge(p, v, w) for p in points
    ])

    expected_result = np.array([2, np.sqrt(18), 1, np.sqrt(2)])

    assert (result == expected_result).all()

def test_compute_distance_to_plane():
    vertices = np.array([[1,0,0], 
                         [4,0,0], 
                         [3,4,0]])
    points = np.array([[3,1,2], [3,-3,0], [10,1,-1]])

    result = np.array([
        vae.compute_distance_to_plane(p, vertices) for p in points
    ])

    expected_result = np.array([2, 0, 1])

    assert (result == expected_result).all()

def test_check_edge_partition():

    vertices = np.array([[1,0,0], 
                         [4,0,0]])
    points = np.array([[1,4,3], [3,-3,0], [0,-1,-1], [6,1,-1], [3,0,0]])

    result = np.array([
        vae.check_edge_partition(p,
                                vertices) for p in points
    ])

    expected_result = np.array([5, 3, np.sqrt(3), np.sqrt(6), 0])

    assert (result == expected_result).all()

def test_vertex_partition():
    v = np.array([0,1,0])
    vert = np.array([[-10,0,0], [10, 0, 0]])
    points = np.array([[0,2,0], [-1,20,2], [-1,2,0], [-21,3,-1]])

    result = np.array([
        vae.check_vertex_partition(p,
                                 v,
                                 vert) for p in points
    ])

    expected_result = np.array([1, np.sqrt(366), 11/np.sqrt(101), np.sqrt(131)])

    assert (result == expected_result).all()

def test_compute_distance_to_simplex():
    vertices = np.array([[1,0,0], 
                         [4,0,0], 
                         [3,4,0]])
    points = np.array([[3,1,2], [0,-1,2], [3,-3,0], [10,1,-1], [3,1,0], [1,0,0]])

    result = np.array([
        vae.compute_distance_to_simplex(p, vertices) for p in points
    ])

    expected_result = np.array([2, np.sqrt(6), 3, np.sqrt(38), 0, 0])

    assert (result == expected_result).all()

def test_compute_distance_to_hull():
    hull_df = pd.read_csv("ref_databases/gcrefv1/convex_hulls.tsv", sep='\t')
    hull = ConvexHull(hull_df.loc[hull_df.species=='Anopheles_coluzzii', ['mean1', 'mean2', \
                                                                          'mean3']].values)
    pos = hull_df.loc[hull_df.species=='Anopheles_gambiae', ['mean1', 'mean2', 'mean3']].values

    result = vae.compute_distance_to_hull(
        hull, 
        pos
    )

    expected_result = np.array([6.309639,13.038807,6.575978,68.123405,7.7456098,24.679714,\
                        72.25647,76.74071,6.5742974,17.787983,66.25579,73.59986,\
                        76.7035,5.73891,63.732883,23.117966,8.014164,18.58697,\
                        30.661522,31.2423,47.31292,49.995487,49.006126,44.459553,\
                        52.336056,23.796885,59.219532,69.760345,26.823551,4.229266,\
                        77.5418,53.13494,5.8358717,8.693414,2.8645864])
    assert (np.abs(result-expected_result) < 0.0001).all()

def test_check_is_in_hull():
    hull_df = pd.read_csv("ref_databases/gcrefv1/convex_hulls.tsv", sep='\t')
    hull_pos = hull_df.loc[hull_df.species=='Anopheles_coluzzii', ['mean1', 'mean2', \
                                                                          'mean3']].values
    positions = np.array([[-84.87537,-4.0789433,-18.389744],
                          [-61.125664,8.855109,-19.726767],
                          [-85, 12, -30],
                          [-74.1326,14.065717,-28.022907],
                          [-52.084938,44.803505,-41.526775],
                          [-6.032356,-62.52318,40.535034]])
    
    result = vae.check_is_in_hull(
        hull_pos, 
        positions)

    assert (result == np.array([False, False, True, True, False, False])).all()

def test_perform_convex_hull_assignments():
    latent_pos_df = pd.read_csv("test_data/comparisons/latent_coordinates.tsv", sep='\t', \
                             index_col=0)
    vae_samples = np.array(['DN806197N_A1', 'DN806197N_A10', 'DN806197N_A11', 'DN806197N_A12', \
                            'DN806197N_A2', 'DN806197N_A3', 'DN806197N_A4', 'DN806197N_A5', \
                            'DN806197N_A6', 'DN806197N_A7', 'DN806197N_A8', 'DN806197N_A9'])
    latent_pos_df = latent_pos_df.query('sample_id in @vae_samples')
    hull_dict = vae.generate_convex_hulls(pd.read_csv("ref_databases/gcrefv1/convex_hulls.tsv", sep='\t'))

    result = vae.perform_convex_hull_assignments(
        hull_dict, 
        latent_pos_df
    )
    assert (result.VAE_species.values == np.array(['Uncertain_coluzzii_tengrela_gambiae', 'Anopheles_coluzzii', \
                'Anopheles_coluzzii', 'Anopheles_coluzzii', 'Anopheles_coluzzii', 'Uncertain_tengrela_coluzzii_gambiae',\
                'Anopheles_coluzzii', 'Uncertain_coluzzii_tengrela', 'Anopheles_coluzzii', \
                'Anopheles_coluzzii', 'Anopheles_coluzzii', 'Anopheles_coluzzii'])).all()



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
import re

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'

import argparse
import keras
from scipy.spatial import ConvexHull, Delaunay

from anospp_analysis.util import *
from anospp_analysis.nn import parse_seqids_series, construct_unique_kmer_table

#Variables
K = 8
LATENTDIM = 3
SEED = 374173
WIDTH = 128
DEPTH = 6
DISTRATIO = 7

def prep_reference_index(reference_path):
    '''
    Read in standardised reference index files from database (currently directory)
    '''

    logging.info(f'importing reference index {reference_path}')

    reference_path = reference_path.rstrip('/')

    reference_version = reference_path.split('/')[-1]

    assert re.match(r'^gcrefv\d', reference_version), f'{reference_version} not recognised as gc vae ref version'

    assert os.path.isdir(reference_path), \
        f'reference version {reference_version} does not exist at {reference_path}'

    assert os.path.isfile(f'{reference_path}/selection_criteria.txt'), \
        f'reference version {reference_version} at {reference_path} does not contain required selection_criteria.txt file'
    selection_criteria_file = f'{reference_path}/selection_criteria.txt'

    assert os.path.isfile(f'{reference_path}/_weights.hdf5'), \
        f'reference version {reference_version} at {reference_path} does not contain required _weights.hdf5 file'
    vae_weights_file = f'{reference_path}/_weights.hdf5'

    assert os.path.isfile(f'{reference_path}/convex_hulls.tsv'), \
        f'reference version {reference_version} at {reference_path} does not contain required convex_hulls.tsv file'
    convex_hulls_df = pd.read_csv(f'{reference_path}/convex_hulls.tsv', sep='\t')

    if not os.path.isfile(f'{reference_path}/colors.tsv'):
        logging.warning('No colors defined for plotting.')
        colorsdict = dict()
    else:
        colors = pd.read_csv(f'{reference_path}/colors.tsv',sep='\t', index_col=0)
        colorsdict = dict(colors.color)

    if not os.path.isfile(f'{reference_path}/latent_coordinates.tsv'):
        logging.warning('No reference coordinates defined for plotting.')
        ref_coord = pd.DataFrame()
    else:
        ref_coord = pd.read_csv(f'{reference_path}/latent_coordinates.tsv',sep='\t')

    if os.path.isfile(f'{reference_path}/version.txt'):
        with open(f'{reference_path}/version.txt', 'r') as fn:
            for line in fn:
                version_name = line.strip()
    else:
        logging.warning(f'No version.txt file present for reference version {reference_version} at {reference_path}')
        version_name = 'unknown'
    if version_name != reference_version:
        logging.warning(f'Reference directory name {reference_version} does not match version.txt: {version_name}')

        
    return (selection_criteria_file, vae_weights_file, convex_hulls_df, colorsdict, ref_coord, version_name)

# def read_selection_criteria(selection_criteria_file, nn_stats_df, nn_hap_df):
    
#     level, sgp, n_targets = open(selection_criteria_file).read().split('\t')
#     return select_samples(nn_stats_df, nn_hap_df, level, sgp, int(n_targets))

def select_samples(selection_criteria_file, nn_stats_df, nn_hap_df):
    '''
    Select the samples meeting the criteria for VAE assignment
    Based on NN assignment and number of targets
    '''
    #identify samples meeting selection criteria
    selection_criteria = pd.read_csv(selection_criteria_file, sep='\t')
    vae_samples = []
    for _, r in selection_criteria.iterrows():
        crit_vae_samples = nn_stats_df.loc[
            (nn_stats_df[f'nn_{r.level}'] == r.sgp) & (nn_stats_df['mosq_targets_recovered'] >= r.n_targets), 
            'sample_id'
            ]
        vae_samples.append(crit_vae_samples)
    vae_samples = pd.concat(vae_samples)
    
    #subset haplotype df
    vae_hap_df = nn_hap_df.query('sample_id in @vae_samples')
    
    logging.info(f'selected {len(vae_samples)} samples to be run through VAE')

    return vae_samples, vae_hap_df

def prep_sample_kmer_table(kmers_unique_seqs, parsed_seqids):
    '''
    Prepare k-mer table for a single sample
    '''
    #set up empty arrays
    total_targets = kmers_unique_seqs.shape[0]
    table = np.zeros((total_targets, kmers_unique_seqs.shape[2]), dtype='int')
    n_haps = np.zeros((total_targets), dtype='int')

    for _, row in parsed_seqids.iterrows():
        #only record the first two haplotypes for each target
        if n_haps[row.target] < 2:
            n_haps[row.target] += 1
            table[row.target,:] += kmers_unique_seqs[row.target, row.uidx, :]
    #double counts for homs
    for target in np.arange(total_targets):
        if n_haps[target] == 1:
            table[target,:] *= 2
    #sum over targets
    summed_table = np.sum(table, axis=0)

    return summed_table

def prep_kmers(vae_hap_df, vae_samples, k):
    '''
    Prepare k-mer table for the samples to be run through VAE
    '''
    #translate unique sequences to k-mers
    kmers_unique_seqs = construct_unique_kmer_table(vae_hap_df, k, source='vae samples')
    
    # logging.info('generating k-mer tables for selected samples')

    #set up k-mer table
    kmers_samples = np.zeros((len(vae_samples), 4**k))

    #fill in table for samples
    for n, sample in enumerate(vae_samples):
        parsed_seqids = parse_seqids_series(
            vae_hap_df.loc[vae_hap_df.sample_id == sample, 'seqid']
            )
        kmers_samples[n,:] = prep_sample_kmer_table(kmers_unique_seqs, parsed_seqids)

    return(kmers_samples)

def latent_space_sampling(args):
    
    #Add noise to encoder output
    z_mean, z_log_var = args
    epsilon = keras.backend.random_normal(
        shape=(keras.backend.shape(z_mean)[0], LATENTDIM),
        mean=0, 
        stddev=1., 
        seed=SEED
        )
    
    return z_mean + keras.backend.exp(z_log_var) * epsilon

def define_vae_input(k):
    input_seq = keras.Input(shape=(4**k,))

    return input_seq

def define_encoder(k):
    input_seq = define_vae_input(k)
    x = keras.layers.Dense(WIDTH, activation = 'elu')(input_seq)
    for i in range(DEPTH - 1):
        x = keras.layers.Dense(WIDTH, activation = 'elu')(x)
    z_mean = keras.layers.Dense(LATENTDIM)(x)
    z_log_var = keras.layers.Dense(LATENTDIM)(x)
    z = keras.layers.Lambda(
        latent_space_sampling, 
        output_shape=(LATENTDIM,),
        name = 'z'
        )([z_mean, z_log_var])
    encoder = keras.models.Model(input_seq, [z_mean, z_log_var, z], name = 'encoder')

    return encoder

def define_decoder(k):
    #Check whether you need the layer part here
    decoder_input = keras.layers.Input(shape=(LATENTDIM,), name='ls_sampling')
    x = keras.layers.Dense(WIDTH, activation='linear')(decoder_input)
    for i in range(DEPTH - 1):
        x = keras.layers.Dense(WIDTH, activation='elu')(x)
    output = keras.layers.Dense(4**k, activation='softplus')(x)
    decoder = keras.models.Model(decoder_input, output, name='decoder')

    return decoder

def define_vae(k):
    input_seq = define_vae_input(k)
    encoder = define_encoder(k)
    decoder = define_decoder(k)
    output_seq = decoder(encoder(input_seq)[2])
    vae = keras.models.Model(input_seq, output_seq, name='vae')

    return vae, encoder

def predict_latent_pos(kmer_table, vae_samples, k, vae_weights_file):

    '''
    Predict latent space of test samples based on reference database
    '''

    logging.info('predicting latent space of test samples based on reference database')

    vae, encoder = define_vae(k)

    vae.load_weights(vae_weights_file)
    predicted_latent_pos = encoder.predict(kmer_table, verbose=False)

    predicted_latent_pos_df = pd.DataFrame(
        index=vae_samples, 
        columns=['mean1', 'mean2', 'mean3', 'sd1', 'sd2', 'sd3']
        )
    for i in range(3):
        predicted_latent_pos_df[f'mean{i + 1}'] = predicted_latent_pos[0][:,i]
        predicted_latent_pos_df[f'sd{i + 1}'] = predicted_latent_pos[1][:,i]

    return predicted_latent_pos_df

def generate_convex_hulls(convex_hulls_df):
    '''
    Read in pre-computed points for convex hulls for each species
    '''
    logging.info('setting up convex hulls from reference database')
    hull_dict = dict()
    for species in convex_hulls_df.species.unique():
        pos = convex_hulls_df.loc[
            convex_hulls_df.species==species, 
            ['mean1', 'mean2', 'mean3']
            ].values
        hull = ConvexHull(pos)
        hull_dict[species] = (pos, hull)

    return hull_dict

def check_half_space(p, n, o):
    '''
    check in which half space separated by the plane
    defined by its normal n p lies
    centered on origin o
    '''
    hp = np.dot(p - o, n)
    
    return hp >= 0

def find_normal_edge_simplex(a, b, c):
    '''
    find the normal vector of the plane C
    spanned by the edge ab of the simplex abc
    and the normal vector to the simplex
    centered on a
    oriented s.t. positive points into the simplex
    '''
    #the normal vector of the plane of the simplex
    #centered on a
    z = np.cross(b - a, c - a)
    #normalised
    u = z / np.sqrt(np.dot(z, z))
    #the edge ab normalised and centered on a
    e = (b - a) / np.sqrt(np.dot(b - a, b - a))
    #the unit normal of plane C
    n = np.cross(u, e)
    
    return n, a

def check_edge_projection(p, v, w):
    '''
    check whether p lies above the edge vw
    '''
    #let vw be the normal vector defining plane U
    #through v
    e = w - v
    #length of vw
    le = np.sqrt(np.dot(e, e))
    #unit vector along the edge, centered on v
    n = e / le
    #project p along the edge vw
    q = np.dot(p - v, n)
    
    return q, le

def check_edge_partition(p, vertices):
    '''
    given that p lies in the part bordering edge vw
    determine whether it is closest to the edge 
    or one of the vertices
    '''
    v, w = vertices
    #get lenght of projection q of p along the edge
    #and the lenght l of the edge
    q, l = check_edge_projection(p, v, w)
    
    if q < 0:
        #vertex v is closest
        return compute_distance_to_vertex(p, v)
    elif q > l:
        #vertex w is closest
        return compute_distance_to_vertex(p, w)
    else:
        #edge vw is closest
        return compute_distance_to_edge(p, v, w)

def check_vertex_partition(p, v, verts):
    '''
    given that p lies in the part bordering vertex v
    determine whether it is closest to the vertex 
    or one of the edges vw or vu
    '''
    w1, w2 = verts
    #get lenght of projection q1 of p along the edge vw1
    #and the lenght l1 of the edge vw1
    q1, l1 = check_edge_projection(p, v, w1)
    if q1 > 0:
        #corner of more than 90 degrees
        if q1 > l1:
            #vertex w1 is closest
            return compute_distance_to_vertex(p, w1)
        else:
            #edge vw1 is closest
            return compute_distance_to_edge(p, v, w1)
    q2, l2 = check_edge_projection(p, v, w2)
    if q2 > 0:
        #corner of more than 90 degrees
        if q2 > l2:
            #vertex w2 is closest
            return compute_distance_to_vertex(p, w2)
        else:
            #edge vw2 is closest
            return compute_distance_to_edge(p, v, w2)
    else:
        #vertex v is closest
        return compute_distance_to_vertex(p, v)

def compute_distance_to_plane(p, vertices):
    '''
    compute distance point p to the plane defined 
    by the vertices of the simplex
    '''
    a, b, c = vertices
    #find a vector perpendicular to the simplex
    #centered on a
    z = np.cross(b - a, c - a)
    #unit vector perpendicular to the plane
    u = z / np.sqrt(np.dot(z, z))
    #distance of p to the plane
    h = np.abs(np.dot(p - a, u))
    
    return h

def compute_distance_to_vertex(p, v):
    '''
    compute distance of point p to vertex v
    '''
    d = np.sqrt(np.dot(p - v, p - v))
    
    return d

def compute_distance_to_edge(p, v, w):
    '''
    compute the distance of p to the edge vw
    '''
    #let vw be the normal vector defining plane U
    #through v
    e = w - v
    #let n be unit vector along the edge
    n = e / np.sqrt(np.dot(e, e))
    #let q be the projection of p along the edge vw
    q = np.dot(p - v, n) * n
    #distance is the length of the difference between p and q
    d = np.sqrt(np.dot(p - v - q, p - v - q))
    
    return d

def compute_distance_to_simplex(p, vertices):
    '''
    Compute the 3d distance of point p to the triangle
    defined by the vertices
    '''
    a, b, c = vertices
    
    #get normal vectors to the planes containing the 
    #edges of the triangles and perpendicular to the simplex
    Cn, Co = find_normal_edge_simplex(a, b, c)
    Bn, Bo = find_normal_edge_simplex(c, a, b)
    An, Ao = find_normal_edge_simplex(b, c, a)
    
    #for each of the edge planes, check whether p lies
    #in the same halfplane of the simplex
    Cs = check_half_space(p, Cn, Co)
    Bs = check_half_space(p, Bn, Bo)
    As = check_half_space(p, An, Ao)
    
    partition = np.array([As, Bs, Cs])
    
    if np.sum(partition) == 3:
        #p lies within the contours of the simplex
        return compute_distance_to_plane(p, vertices)
    elif np.sum(partition) == 2:
        #p lies in a part bordering an edge
        return check_edge_partition(p, vertices[partition])
    elif np.sum(partition) == 1:
        #p lies in a part bordering a vertex
        return check_vertex_partition(p, vertices[partition][0], vertices[~partition])
    else:
        logging.error(
            'Something thought to be impossible happened in the convex hull computation - '
            'tell Marilou to retake her linear algebra exam'
            )

def compute_distance_to_hull(hull, positions):
    '''
    compute the distance of all given positions
    to the given convex hull
    in 3d
    '''
    #array to store distances to all simplices
    dists = np.zeros((positions.shape[0], hull.simplices.shape[0]))
    for i, s in enumerate(hull.simplices):
        #compute distance to each simplex in turn
        dists[:, i] = np.array([compute_distance_to_simplex(p, hull.points[s]) for p in positions])
    #get the distance to the closest simplex for each point
    distances = dists.min(axis=1)

    return distances

def check_is_in_hull(hull_pos, positions):
    '''
    Check whether a set of positions lies inside the specified hull
    '''
    if not isinstance(hull_pos,Delaunay):
        hull = Delaunay(hull_pos)

    in_hull = hull.find_simplex(positions) >= 0

    return in_hull

def get_unassigned_samples(latent_positions_df):
    unassigned = latent_positions_df.loc[latent_positions_df.vae_species_call.isnull()].index
    n_unassigned = len(unassigned)
    return unassigned, n_unassigned

def generate_hull_dist_df(hull_dict, latent_positions_df, unassigned):

    dist_df = pd.DataFrame(index=unassigned)
    positions = latent_positions_df.loc[unassigned ,['mean1', 'mean2', 'mean3']].values
    for species in hull_dict.keys():
        dist_df[species] = compute_distance_to_hull(hull_dict[species][1], positions)
    return dist_df 

def get_closest_hulls(hull_dict, latent_positions_df, unassigned):

    dist_df = generate_hull_dist_df(hull_dict, latent_positions_df, unassigned)
    summary_dist_df = pd.DataFrame(index=dist_df.index)
    summary_dist_df['dist1'] = dist_df.min(axis=1)
    summary_dist_df['species1'] = dist_df.idxmin(axis=1)
    summary_dist_df['dist2'] = dist_df.apply(lambda x: x.sort_values().iloc[1], axis=1)
    summary_dist_df['species2'] = dist_df.apply(lambda x: x.sort_values().index[1], axis=1)
    return summary_dist_df, dist_df

def assign_gam_col_band(latent_positions_df, summary_dist_df):
    #Determine which samples are in gamcol band
    gamcol_band = summary_dist_df.loc[
        (summary_dist_df.species1.isin(['Anopheles_gambiae', 'Anopheles_coluzzii'])) & \
        (summary_dist_df.species2.isin(['Anopheles_gambiae', 'Anopheles_coluzzii'])) & \
        (summary_dist_df.dist2 < 14)
        ].copy()
    #Make assignments for the samples in gamcol band
    if gamcol_band.shape[0] > 0:
        gamcol_dict = dict(gamcol_band.apply(
            lambda row: 'Uncertain_' + row.species1.split('_')[1] + '_' + row.species2.split('_')[1],
            axis=1))
        latent_positions_df.loc[gamcol_band.index, 'vae_species_call'] = latent_positions_df.loc[\
            gamcol_band.index
            ].index.map(gamcol_dict)
    return latent_positions_df, gamcol_band.shape[0]

def assign_to_closest_hull(latent_positions_df, summary_dist_df, unassigned):
    '''
    Assign samples that are much closer to one hull than to all others to that hull
    Currently defined as a distance 7 times smaller than all others
    '''
    fuzzy_hulls = summary_dist_df.loc[
        (DISTRATIO*summary_dist_df.dist1 < summary_dist_df.dist2) & \
        summary_dist_df.index.isin(unassigned)
        ].copy()
    if fuzzy_hulls.shape[0] > 0:
        fuzzy_hulls_dict = dict(fuzzy_hulls.species1)
        latent_positions_df.loc[unassigned, 'vae_species_call'] = latent_positions_df.loc[
            unassigned
            ].index.map(fuzzy_hulls_dict)
    return latent_positions_df, fuzzy_hulls.shape[0]

def assign_to_multiple_hulls(latent_positions_df, dist_df, unassigned):
    '''
    Assign samples in between hulls to multiple hulls in order of closeness
    '''
    between_hulls = dist_df.loc[dist_df.index.isin(unassigned)].copy()
    between_hulls_dict = dict(between_hulls.apply(
        lambda row: 'Uncertain_'+'_'.join(
            label.split('_')[1] for label in row.sort_values().index[
                row.sort_values() < DISTRATIO * row.min()
                ]
            ),
        axis=1))
    latent_positions_df.loc[unassigned, 'vae_species_call'] = latent_positions_df.loc[
        unassigned
        ].index.map(between_hulls_dict)
    return latent_positions_df, between_hulls.shape[0]

def perform_convex_hull_assignments(hull_dict, latent_positions_df):
    '''
    Perform convex hull assignments based on latent space positions
    '''
    logging.info('performing convex hull assignment')
    positions = latent_positions_df[['mean1', 'mean2', 'mean3']].values
    
    #first check which samples fall inside convex hulls
    for label in hull_dict.keys():
        is_in_hull = check_is_in_hull(hull_dict[label][0], positions)
        latent_positions_df.loc[is_in_hull, 'vae_species_call'] = label
    
    #Record unassigned samples 
    unassigned, n_unassigned = get_unassigned_samples(latent_positions_df)
    logging.info(
        f'{latent_positions_df.shape[0] - n_unassigned} samples fall inside convex hulls; '
        f'{n_unassigned} samples still to be assigned'
        )
    
    #for the unassigned samples, get distances to two closest hulls
    if n_unassigned > 0:
        summary_dist_df, dist_df = get_closest_hulls(hull_dict, latent_positions_df, unassigned)
        latent_positions_df, n_newly_assigned = assign_gam_col_band(
            latent_positions_df, 
            summary_dist_df
            )
        unassigned, n_unassigned = get_unassigned_samples(latent_positions_df)
        logging.info(
            f'{n_newly_assigned} samples assigned to uncertain_gambiae_coluzzii or uncertain_coluzzii_gambiae; '
            f'{n_unassigned} samples still to be assigned'
            )
    if n_unassigned > 0:
        latent_positions_df, n_newly_assigned = assign_to_closest_hull(
            latent_positions_df, 
            summary_dist_df, 
            unassigned
            )
        unassigned, n_unassigned = get_unassigned_samples(latent_positions_df)
        logging.info(
            f'{n_newly_assigned} additional samples assigned to a single species; '
            f'{n_unassigned} samples still to be assigned')
    if n_unassigned > 0:
        latent_positions_df, n_newly_assigned = assign_to_multiple_hulls(
            latent_positions_df,
            dist_df, 
            unassigned
            )
        logging.info(f'{n_newly_assigned} samples assigned to multiple hulls.')
    logging.info(f'finished assigning {latent_positions_df.shape[0]} samples')

    return latent_positions_df

def plot_vae_assignments(ch_assignments, ref_coordinates, colordict, run_id):

    logging.info('generating VAE plots')
    #get assignment categories
    single, double, multi = [], [], []
    species_dict = dict(ch_assignments.groupby('vae_species_call').size())
    for key in species_dict.keys():
        #for single species assignment
        if key.startswith('Anopheles_'):
            single.append(key)
        #for uncertain between two species
        elif len(key.split('_')) == 3:
            double.append(key)
        #for uncertain between more than two species
        else:
            multi.append(key)

    fig, ax = plt.subplots(figsize=(8, 8), constrained_layout=True)

    fig.suptitle(f'An. gambiae complex VAE assignment for run {run_id}')

    #Plot reference dataset
    ref_coordinates['color'] = ref_coordinates.vae_species_call.map(colordict)
    ax.scatter(
        ref_coordinates.mean1, 
        ref_coordinates.mean3, 
        c=ref_coordinates.color.values, 
        alpha=.6, 
        zorder=1
        )
    
    #Plot samples assigned to single species
    if len(single) > 0:
        for species in single:
            sub = ch_assignments.query('vae_species_call == @species')
            if species == 'Anopheles_bwambae-fontenillei':
                label = f'Anopheles_bwambae-\nfontenillei: {species_dict[species]}'
            else:
                label = f'{species}: {species_dict[species]}'
            ax.scatter(
                sub.mean1, 
                sub.mean3, 
                color=colordict[species], 
                marker='^', 
                edgecolor='k', 
                label=label, 
                zorder=2
                )

    #Plot samples assigned to two species
    if len(double)>0:    
        for species in double:
            sub = ch_assignments.query('vae_species_call == @species')
            _, sp1, sp2 = species.split('_')
            #multiple lines on the legend
            if sp1 == 'bwambae-fontenillei':
                label = f'Uncertain_bwambae-\nfontenillei_{sp2}: {species_dict[species]}'
            else:
                label = f'Uncertain_{sp1}_\n{sp2}: {species_dict[species]}'
            ax.scatter(
                sub.mean1, 
                sub.mean3, 
                marker='s', 
                edgecolor=colordict[f'Anopheles_{sp1}'], 
                facecolor=colordict[f'Anopheles_{sp2}'], 
                linewidth=1.5, 
                zorder=2, 
                label=label
                )

    #Plot remaining samples
    if len(multi)>0:    
        sub = ch_assignments.query('vae_species_call in @multi')
        ax.scatter(
            sub.mean1, 
            sub.mean3,
            color='k', 
            marker='s', 
            zorder=3, 
            label=f'other: {sub.shape[0]}'
            )

    ax.set_xlabel('LS1')
    ax.set_ylabel('LS3')
        
    legend = ax.legend(loc='lower right', prop={'size': 9})

    return fig, ax

def generate_summary(ch_assignments, nn_stats_df, version_name):
    summary = [
        f'Convex hull assignment assignment using reference version {version_name}',
        f'On run containing {nn_stats_df.sample_id.nunique()} samples',
        f'{ch_assignments.index.nunique()} samples met the VAE selection criteria',
        f'Samples are assigned to the following labels:'
    ]
    assignment_dict = dict(ch_assignments.groupby('vae_species_call').size())
    for key in assignment_dict.keys():
        summary.append(f'{key}:\t {assignment_dict[key]}')

    return '\n'.join(summary)


def vae(args):

    setup_logging(verbose=args.verbose)

    os.makedirs(args.outdir, exist_ok =True)

    logging.info('ANOSPP VAE data import started')

    nn_hap_df = pd.read_csv(args.nn_haplotypes, sep='\t')
    nn_stats_df = pd.read_csv(args.nn_manifest, sep='\t')
    run_id = nn_stats_df['run_id'].iloc[0]

    selection_criteria_file, vae_weights_file, convex_hulls_df, colordict, ref_coordinates, \
        version_name = prep_reference_index(args.reference_path)
    vae_samples, vae_hap_df = select_samples(
        selection_criteria_file,
        nn_stats_df,
        nn_hap_df
        )
    if len(vae_samples) == 0:
        logging.info('No samples to be run through VAE - skipping to finalising assignments')
        ch_assignment_df = pd.DataFrame(columns = [
            'mean1', 'mean2', 'mean3', 'sd1', 'sd2', 'sd3', 'vae_species_call'
            ])
        ch_assignment_df.index.name = 'sample_id'

    else:
        kmer_table = prep_kmers(vae_hap_df, vae_samples, K)
        latent_positions_df = predict_latent_pos(kmer_table, vae_samples, K, vae_weights_file)
        hull_dict = generate_convex_hulls(convex_hulls_df)
        ch_assignment_df = perform_convex_hull_assignments(hull_dict, latent_positions_df)
        if not args.no_plotting:
            fig, _ = plot_vae_assignments(ch_assignment_df, ref_coordinates, colordict, run_id)
            fig.savefig(f'{args.outdir}/vae_assignment.png')

    ch_assignment_df['vae_ref'] = version_name
    logging.warning('Changing VAE species predictions prefix from "Anopheles_" to "An_"')
    ch_assignment_df['vae_species_call'] = ch_assignment_df['vae_species_call'] \
        .str.replace('^Anopheles_','An_', regex=True)
    ch_assignment_df.to_csv(f'{args.outdir}/vae_assignment.tsv', sep='\t')

    summary_text = generate_summary(ch_assignment_df, nn_stats_df, version_name)
    logging.info(f'writing summary file to {args.outdir}')
    with open(f'{args.outdir}/summary.txt', 'w') as fn:
        fn.write(summary_text)

    logging.info('ANOSPP VAE complete')

    
def main():
    
    parser = argparse.ArgumentParser('VAE assignment for samples in Anopheles gambiae complex')
    parser.add_argument('-a', '--nn_haplotypes',
                        help=f'nn_hap_summary.tsv generated by anospp-nn with k={K}',
                        required=True)
    parser.add_argument('-m', '--nn_manifest',
                        help=f'nn_assignment.tsv file generated by anospp-nn with k={K}',
                        required=True)
    parser.add_argument('-r', '--reference_path', 
                        help='Path to vae reference index directory',
                        required=True)
    parser.add_argument('-o', '--outdir', help='Output directory. Default: vae', default='vae')
    parser.add_argument('--no_plotting', 
                        help='Do not generate plots. Default: False',
                        action='store_true', default=False)
    parser.add_argument('-v', '--verbose', 
                        help='Include INFO level log messages', action='store_true')

    args = parser.parse_args()
    args.outdir=args.outdir.rstrip('/')

    vae(args)


if __name__ == '__main__':
    main()


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import os
import re
import argparse
import itertools

from anospp_analysis.util import *
from anospp_analysis.qc import plot_het_cov

def prep_mosquito_haps(hap_df, rc_threshold, rf_threshold):
    '''
    prepare mosquito haplotype dataframe
    remove plasmodium haplotypes
    change targets to integers
    returns haplotype dataframe
    '''

    logging.info('preparing mosquito haplotypes')

    hap_df = hap_df.astype({'target': str})
    filtered_hap_df = hap_df[
        (hap_df.reads >= rc_threshold) & (hap_df.reads_fraction >= rf_threshold)
        ]
    if filtered_hap_df.shape[0] < hap_df.shape[0]:
        logging.info(
            f'removed {hap_df.shape[0] - filtered_hap_df.shape[0]} haplotypes '
            f'with fewer than {rc_threshold} reads or fraction lower than {rf_threshold} of reads'
        )
    assert filtered_hap_df.shape[0] > 0, 'No haplotypes left after filtering, terminating'
    mosq_hap_df = filtered_hap_df[filtered_hap_df.target.isin(MOSQ_TARGETS)]
    mosq_hap_df = mosq_hap_df.astype({'target': int})

    #recompute reads coverage after filtering
    mosq_hap_df = recompute_haplotype_coverage(mosq_hap_df)

    return mosq_hap_df

def prep_reference_index(reference_path):
    '''
    Read in standardised reference index files from database (currently directory)
    '''

    logging.info(f'importing reference index from {reference_path}')

    reference_path = reference_path.rstrip('/')

    reference_version = reference_path.split('/')[-1]

    assert re.match(r'^nnv\d', reference_version), f'{reference_version} not recognised as NN ref version'

    assert os.path.isdir(reference_path), f'reference version {reference_version} does not exist at {reference_path}'

    assert os.path.isfile(f'{reference_path}/haplotypes.tsv'), \
        f'reference version {reference_version} at {reference_path} does not contain required haplotypes.tsv file'
    ref_hap_df = pd.read_csv(f'{reference_path}/haplotypes.tsv', sep='\t')

    assert os.path.isfile(f'{reference_path}/multiallelism.tsv'), \
        f'reference version {reference_version} at {reference_path} does not contain required multiallelism.tsv file'
    true_multi_targets = pd.read_csv(f'{reference_path}/multiallelism.tsv', sep='\t')

    if os.path.isfile(f'{reference_path}/version.txt'):
        with open(f'{reference_path}/version.txt', 'r') as fn:
            for line in fn:
                version_name = line.strip()
    else:
        logging.warning('No version.txt file present for reference version {reference_version} at {reference_path}')
        version_name = 'unknown'
    if version_name != reference_version:
        logging.warning(f'Reference directory name {reference_version} does not match version.txt: {version_name}')

    allele_freqs = dict()
    colors = dict()
    for level in ['coarse', 'int', 'fine']:
        assert os.path.isfile(f'{reference_path}/allele_freq_{level}.npy'), \
            f'reference version {reference_version} at {reference_path} does not contain required allele_freq_{level}.npy file'
        af = np.load(f'{reference_path}/allele_freq_{level}.npy')
        allele_freqs[level] = af

        assert os.path.isfile(f'{reference_path}/sgp_{level}.txt'), \
            f'reference version {reference_version} at {reference_path} does not contain required sgp_{level}.txt file'
        sgp = []
        with open(f'{reference_path}/sgp_{level}.txt', 'r') as fn:
            for line in fn:
                sgp.append(line.strip())

        ref_hap_df[f'{level}_sgp'] = pd.Categorical(ref_hap_df[f'{level}_sgp'], sgp, ordered=True)
        assert ~ref_hap_df[f'{level}_sgp'].isna().any(), \
            f'reference version {reference_version} at {reference_path} has inconsistent labels between haplotypes.tsv and sgp_{level}.txt'

        if not os.path.isfile(f'{reference_path}/colors_{level}.npy'):
            logging.warning('No colors defined for plotting.')
        else:
            clr = np.load(f'{reference_path}/colors_{level}.npy')
            colors[level] = clr
        
    return(ref_hap_df, allele_freqs, true_multi_targets, colors, version_name)

def construct_kmer_dict(k):
    '''
    construct a k-mer dict
    associating each unique k-mer of length k with a unique non-negative integer <4**k
    bases are written in capitals
    returns a dictionary
    '''
    labels = []
    for i in itertools.product('ACGT', repeat=k):
        labels.append(''.join(i))
    kmerdict = dict(zip(labels, np.arange(4**k)))
    return kmerdict   

def parse_seqid(seqid):
    '''
    Parse seqid passed as a string 
    '''
    split_seqid = seqid.split('-')

    assert split_seqid[0] in MOSQ_TARGETS, f'seqid {seqid} refers to a non-mosquito target'
    try:
        parsed_seqid = (int(split_seqid[0]), int(split_seqid[1]))
    except:
        raise Exception(f'seqid {seqid} cannot be converted to integers')
    return parsed_seqid

def parse_seqids_series(seqids):
    '''
    Parse seqid or seqids passed as a pandas Series
    '''
    parsed_seqids = seqids.str.split('-', expand=True)
    # assert parsed_seqids.shape[1] == 2, 'Failed to parse seqids, we expected only one "-" in each seqid'
    parsed_seqids.columns = ['target', 'uidx']
    
    assert parsed_seqids['target'].isin(MOSQ_TARGETS).all(), 'Dataframe contains seqids referring to non-mosquito targets'
    try:
        parsed_seqids = parsed_seqids.astype(int)
    except:
        raise Exception('Dataframe contains seqids which cannot be converted to integers')
    return parsed_seqids

def construct_unique_kmer_table(mosq_hap_df, k, source):
    '''
    constructs a k-mer table of dimensions n_amp * maxallele * 4**k
    represting the k-mer table of each unique sequence in the dataframe
    maxallele is the maximum number of unique sequences per target
    n_amp is the number of mosquito targets
    input: k=k (length of k-mers), hap_df=dataframe with haplotypes
    output: k-mer table representing each unique haplotype in the hap dataframe
    '''

    logging.info(f'translating unique sequences from {source} to {k}-mers')

    kmerdict = construct_kmer_dict(k)
    #subset to unique haplotypes
    uniqueseq = mosq_hap_df[['seqid', 'consensus']].drop_duplicates()
    #determine shape of table by top seqid
    parsed_seqids = parse_seqids_series(uniqueseq.seqid)
    maxid = parsed_seqids['uidx'].max() + 1

    #initiate table to store kmer counts
    kmer_table = np.zeros((len(MOSQ_TARGETS), maxid, 4**k), dtype='int')
    #translate each unique haplotype to kmer counts
    for idx, seq in uniqueseq.iterrows():
        tgt = parsed_seqids.loc[idx, 'target']
        uid = parsed_seqids.loc[idx, 'uidx']
        consensus = seq.consensus
        for i in np.arange(len(consensus) - (k - 1)):
            kmer_table[tgt, uid, kmerdict[consensus[i:i+k]]] += 1
    return kmer_table

def identify_error_seqs(mosq_hap_df, kmers, k, n_error_snps):
    '''
    Identify haplotypes resulting from sequencing/PCR errors
    Cannot distinguish between true heterozygote, contaminated homozygote and homozygote with error sequence
    So only look for errors for unique sequences at multiallelic targets
    '''

    logging.info('identifying haplotypes resulting from sequencing/PCR errors')
    #set the k-mer threshold for the number of snps allowed for errors
    threshold = n_error_snps * k + 1
    seqid_size = mosq_hap_df.groupby('seqid').size()
    singleton_seqids = seqid_size[seqid_size == 1].index # used in query
    error_candidates = mosq_hap_df.query('(seqid in @singleton_seqids) & (nalleles > 2)')

    error_seqs = []
    for _, cand in error_candidates.iterrows():
        possible_sources = mosq_hap_df.query(
            '(sample_id == @cand.sample_id) & (target == @cand.target) & (seqid != @cand.seqid)')
        cand_parsed_seqid = parse_seqid(cand.seqid)
        possible_sources_parsed_seqids = parse_seqids_series(possible_sources.seqid)
        for possible_source in possible_sources_parsed_seqids['uidx']:
            abs_kmer_dist = np.abs(
                kmers[cand.target, cand_parsed_seqid[1], :] - kmers[cand.target, possible_source, :]
                ).sum()
            if abs_kmer_dist < threshold:
                error_seqs.append(cand.seqid)
                break
    
    logging.info(f'identified {len(error_seqs)} error sequences')

    return error_seqs

def recompute_haplotype_coverage(hap_df):
    hap_df = hap_df.drop(['total_reads', 'reads_fraction', 'nalleles'], axis=1)

    hap_df['total_reads'] = hap_df.groupby(by=['sample_id', 'target']) \
            ['reads'].transform('sum')

    hap_df['reads_fraction'] = hap_df['reads'] / hap_df['total_reads']

    hap_df['nalleles'] = hap_df.groupby(by=['sample_id', 'target']) \
            ['consensus'].transform('nunique')
    return hap_df

def compute_kmer_distance(kmers, ref_kmers, tgt, qidx, refidx):
    '''
    compute k-mer distance between query kmer count
    and ref kmer count(s)
    returns absolute and normalised distance
    '''
    #identify k-mer mismatches
    diff = np.abs(ref_kmers[tgt, refidx, :] - kmers[tgt, qidx, :])
    total = np.sum(ref_kmers[tgt, refidx, :] + kmers[tgt, qidx, :], axis=1)
    #sum to get absolute distance
    dist = np.sum(diff, axis=1)
    #normalise
    norm_dist = dist / total

    return dist, norm_dist

def find_nn_unique_haps(non_error_hap_df, kmers, ref_hap_df, ref_kmers):
    '''
    identify the nearest neighbours of the unique haplotypes in the reference dataset 
    '''
    #get idxs occupied for each target
    parsed_ref_seqids = parse_seqids_series(ref_hap_df.seqid.drop_duplicates())
    ref_idxs_per_target = parsed_ref_seqids.groupby('target')['uidx'].unique()

    nndict = dict()

    logging.info(f'identifying nearest neighbours for {non_error_hap_df.seqid.nunique()} unique haplotypes')
    
    #loop through unique haplotypes
    unique_seqids = non_error_hap_df.seqid.unique()
    for seqid in unique_seqids:
        tgt, qidx = parse_seqid(seqid)
        #compute distance between focal hap and all same target haps in ref index
        dist, norm_dist = compute_kmer_distance(kmers, ref_kmers, tgt, qidx, ref_idxs_per_target[tgt])
        #Find nearest neighbours
        nn_qidx = ref_idxs_per_target[tgt][norm_dist==norm_dist.min()]
        #include in dict
        nndict[seqid] = (nn_qidx, norm_dist.min())

    return nndict

def lookup_assignment_proportion(q_seqid, allele_frequencies, tgt, nndict, weight=1):

    #lookup nearest neighbour identifiers
    nnids = nndict[q_seqid][0]
    #lookup allele frequencies of nnids
    af_nn = allele_frequencies[tgt, nnids, :]
    #sum allele frequencies over nnids
    summed_af_nn = np.sum(af_nn, axis=0)
    #normalise proportion and weigth in number of alleles
    assignment_proportion = weight * summed_af_nn / np.sum(summed_af_nn)
    return assignment_proportion

def add_nn_to_haplotypes(non_error_hap_df, nndict, ref_hap_df):

    def get_unique_sgps(x, level, ref_spp):
        seqids = x.split(';')
        match_sgps = ref_spp.loc[seqids, f'{level}_sgp'].transform(set)
        sgps = set().union(*match_sgps)# - set([np.nan])
        if len(sgps) == 0:
            return ''
        return ';'.join(sgps)

    logging.info('adding NN annotations to non-error haplotypes')

    non_error_hap_df['ref_nn_dist'] = non_error_hap_df['seqid'].apply(lambda x: nndict[x][1])
    non_error_hap_df['ref_nn_seqids'] = non_error_hap_df['seqid'].apply(
        lambda x: ";".join([x.split('-')[0] + '-' + str(y) for y in nndict[x][0]])
    )

    ref_spp = ref_hap_df.groupby('seqid')[['fine_sgp','int_sgp','coarse_sgp']].agg(set)

    non_error_hap_df['ref_coarse_sgps'] = non_error_hap_df['ref_nn_seqids'].apply(lambda x: get_unique_sgps(x, 'coarse', ref_spp))
    non_error_hap_df['ref_int_sgps'] = non_error_hap_df['ref_nn_seqids'].apply(lambda x: get_unique_sgps(x, 'int', ref_spp))
    non_error_hap_df['ref_fine_sgps'] = non_error_hap_df['ref_nn_seqids'].apply(lambda x: get_unique_sgps(x, 'fine', ref_spp))
    
    return non_error_hap_df

def perform_nn_assignment_samples(hap_df, ref_hap_df, nndict, allele_freqs, normalisation):
    '''
    The main NN assignment function
    it outputs three dataframes containing the assignment proportions to each species-group for the three levels
    '''
    #get samples with at least 10 targets
    test_samples = hap_df \
        .groupby('sample_id') \
        .filter(lambda x: x['target'].nunique() >= 10)['sample_id'] \
        .unique()

    logging.info(f'performing NN assignment for {len(test_samples)} samples with >=10 mosquito targets')

    #set up data-output as numpy arrays (will be made into dataframes later)
    results = {
        'coarse': np.zeros((len(MOSQ_TARGETS), len(test_samples), allele_freqs['coarse'].shape[2])),
        'int': np.zeros((len(MOSQ_TARGETS), len(test_samples), allele_freqs['int'].shape[2])),
        'fine': np.zeros((len(MOSQ_TARGETS), len(test_samples), allele_freqs['fine'].shape[2]))
    }

    for i, sample in enumerate(test_samples):
        #Restrict to targets amplified in focal sample
        targets = hap_df.loc[hap_df.sample_id == sample, 'target'].unique()
        
        #Per amplified target
        for tgt in targets:
            #Identify the unique IDs of the focal sample's haplotypes at target t
            alleles = hap_df.loc[
                (hap_df.sample_id == sample) & (hap_df.target == tgt),
                ['seqid', 'reads_fraction']]
            #for each haplotype
            for _, allele in alleles.iterrows():
                #for each assignment level
                for level in ['coarse', 'int', 'fine']:
                    if normalisation == 'n_alleles':
                    #lookup assignment proportion
                        assignment_proportion = lookup_assignment_proportion(
                            allele.seqid,
                            allele_freqs[level],
                            tgt,
                            nndict,
                            1 / alleles.shape[0]
                            )
                    elif normalisation == 'reads_fraction':
                        assignment_proportion = lookup_assignment_proportion(
                            allele.seqid,
                            allele_freqs[level],
                            tgt,
                            nndict,
                            allele.reads_fraction
                        )
                    else:
                        logging.error('Not a valid allelism_normalisation method.')
                    #table[tgt,nsmp,:] += assignment_proportion
                    results[level][tgt, i, :] += assignment_proportion

    #print(f'shape of results arrays is {results_coarse.shape}, {results_int.shape} and {results_fine.shape}')
    results_dfs = dict()
    for level in ['coarse', 'int', 'fine']:
        #Average assignment results over amplified targets
        res = np.nansum(results[level], axis=0)/np.sum(np.nansum(results[level], axis=0), axis=1)[:,None]
        #Convert results to dataframes
        results_dfs[level] = pd.DataFrame(
            res, 
            index=test_samples, 
            columns=ref_hap_df[f'{level}_sgp'].cat.categories
        )  
    return results_dfs, test_samples

def recompute_sample_coverage(comb_stats_df, non_error_hap_df):
    '''
    recompute coverage stats after filtering and error removal
    '''
    logging.info('recompute coverage stats')
    comb_stats_df.set_index('sample_id', inplace=True)

    #recompute multiallelic calls after filtering and error removal
    comb_stats_df['multiallelic_mosq_targets'] = (
        non_error_hap_df.groupby('sample_id')['target'].value_counts() > 2
        ).groupby(level='sample_id').sum()

    #recompute read counts after filtering and error removal
    comb_stats_df['mosq_reads'] = non_error_hap_df.groupby('sample_id')['reads'].sum()

    #recompute targets recovered after filtering and error removal
    comb_stats_df['mosq_targets_recovered'] = non_error_hap_df.groupby('sample_id')['target'].nunique()

    for col in ['multiallelic_mosq_targets', 'mosq_reads', 'mosq_targets_recovered']:
        comb_stats_df[col] = comb_stats_df[col].fillna(0).astype(int)

    comb_stats_df.reset_index(inplace=True)

    return comb_stats_df

def generate_hard_calls(comb_stats_df, non_error_hap_df, test_samples, results_dfs,
                        true_multi_targets, nn_asgn_threshold):

    logging.info('generating NN calls from assignment info')

    #Account for filtering and error removal
    comb_stats_df = recompute_sample_coverage(comb_stats_df, non_error_hap_df)

    #Record whether NN assignment was performed
    comb_stats_df.loc[comb_stats_df.sample_id.isin(test_samples), 'nn_assignment'] = 'yes'
    comb_stats_df.loc[comb_stats_df.nn_assignment.isnull(), 'nn_assignment'] = 'no'

    #Generate assignment hard calls if the threshold is met
    for level in ['coarse', 'int', 'fine']:
        asgn_dict = dict(results_dfs[level] \
            .loc[(results_dfs[level] >= nn_asgn_threshold).any(axis=1)] \
            .apply(
                lambda row: results_dfs[level].columns[row >= nn_asgn_threshold][0],
                axis=1))
        comb_stats_df[f'nn_{level}'] = comb_stats_df.sample_id.map(asgn_dict)

    comb_stats_df['nn_species_call'] = None
    comb_stats_df['nn_call_method'] = None
    # NN hierarchical assignment by level
    for level in ['fine', 'int', 'coarse']:
        leveldict = dict(zip(comb_stats_df.sample_id, comb_stats_df[f'nn_{level}']))
        is_id_on_level = (comb_stats_df.nn_species_call.isnull() & ~comb_stats_df[f'nn_{level}'].isnull())
        comb_stats_df.loc[is_id_on_level, 'nn_call_method'] = f'NN_{level}'
        comb_stats_df.loc[is_id_on_level, 'nn_species_call'] = comb_stats_df.loc[
            comb_stats_df.nn_call_method == f'NN_{level}', 'sample_id'
            ].map(leveldict)
    
    #Rainbow samples
    is_rainbow = (comb_stats_df.nn_species_call.isnull() & (comb_stats_df.nn_assignment == 'yes'))
    comb_stats_df.loc[is_rainbow, 'nn_call_method'] = 'NN'
    comb_stats_df.loc[is_rainbow, 'nn_species_call'] = 'RAINBOW_SAMPLE'

    #Samples with too few targets
    is_not_id = (comb_stats_df.nn_assignment == 'no')
    comb_stats_df.loc[is_not_id, 'nn_call_method'] = 'TOO_FEW_TARGETS'
    comb_stats_df.loc[is_not_id, 'nn_species_call'] = 'TOO_FEW_TARGETS'

    assert not comb_stats_df.nn_species_call.isnull().any(), 'some samples not assigned'
    assert not comb_stats_df.nn_call_method.isnull().any(), 'some samples not assigned'

    return comb_stats_df

def generate_summary(comb_stats_df, version_name):

    summary = [
        f'Nearest Neighbour assignment using reference version {version_name}',
        f'On run containing {comb_stats_df.sample_id.nunique()} samples',
        f'{(comb_stats_df.nn_assignment=="no").sum()} samples with < 10 targets lack NN assignment',
        f'{(~comb_stats_df.nn_coarse.isnull()).sum()} samples are assigned at coarse level',
        f'to {comb_stats_df.nn_coarse.nunique()} different species groups',
        f'{(~comb_stats_df.nn_int.isnull()).sum()} samples are assigned at intermediate level',
        f'to {comb_stats_df.nn_int.nunique()} different species groups',
        f'{(~comb_stats_df.nn_fine.isnull()).sum()} samples are assigned at fine level',
        f'to {comb_stats_df.nn_fine.nunique()} different species groups',
        f'{comb_stats_df.loc[comb_stats_df.nn_call_method == "NN_int", "sample_id"].nunique()} '
        f'samples with sufficient coverage could not be assigned at fine level',
        f'{comb_stats_df.loc[comb_stats_df.nn_call_method == "NN_coarse", "sample_id"].nunique()} '
        f'samples with sufficient coverage could not be assigned at intermediate level',
    ]
    return '\n'.join(summary)

def prep_stats_for_plotting(comb_stats_df, locov_rc):

    

    return comb_stats_df

def plot_assignment_proportions(comb_stats_df, nn_level_result_df, level_label, level_colors, run_id, plasm_assignment_df, plasm_colors, args):
    
    logging.info(f'generating {level_label} level plots')

    assert nn_level_result_df.index.isin(comb_stats_df['sample_id']).all(), \
        f'NN {level_label} samples do not match comb stats'
    
    assert nn_level_result_df.index.is_unique, \
        f'NN {level_label} samples are not unique'

    #Generate bar plots at given assignment level
    #add samples with <10 targets
    nn_level_result_df = pd.concat([
        nn_level_result_df, pd.DataFrame(
            index=comb_stats_df.loc[
                ~comb_stats_df.sample_id.isin(nn_level_result_df.index), 'sample_id'
                ]
            )
        ]).fillna(0)

    #Get row and col info from well_id for ordering samples
    comb_stats_df['row_id'] = comb_stats_df.well_id.str[0]
    comb_stats_df['col_id'] = comb_stats_df.well_id.str[1:].astype(int)
    comb_stats_df['well_id'] = well_ordering(comb_stats_df['well_id'])
    # comb_stats_df.sort_values(by=['plate_id', 'col_id', 'well_id'], inplace=True)
    
    # ma bands rely on multiallelics being numeric
    comb_stats_df['multiallelic_mosq_targets'] = comb_stats_df['multiallelic_mosq_targets'].fillna(0).astype(int)
    # multiallelics bands
    comb_stats_df['ma_band'] = '0'
    comb_stats_df.loc[comb_stats_df['multiallelic_mosq_targets'] > 0, 'ma_band'] = '1-2'
    comb_stats_df.loc[comb_stats_df['multiallelic_mosq_targets'] > 2, 'ma_band'] = '3-4'
    comb_stats_df.loc[comb_stats_df['multiallelic_mosq_targets'] > 4, 'ma_band'] = '5+'

    # locov relies on mosq_reads being numeric
    comb_stats_df['mosq_reads'] = comb_stats_df['mosq_reads'].fillna(0).astype(int)
    # mark low coverage samples - this can only be done once
    comb_stats_df['mosq_targets_plot'] = comb_stats_df['mosq_targets_recovered'].astype(str)
    comb_stats_df.loc[
        comb_stats_df['mosq_reads'] < args.locov_rc,
        'mosq_targets_plot'
    ] = comb_stats_df['mosq_targets_plot'] + '*'

    # multiallelics color scheme - applied to top ticks 
    ma_colors = {
        '0':'#808080',
        '1-2':'#FF9900',
        '3-4':'#cc00FF',
        '5+':'#FF0000'
    }

    # plasm color scheme - applied to bottom ticks
    if plasm_assignment_df is not None and plasm_colors is not None:
        logging.info(f'using plasm predictions to colour sample labels')
        plasm_assignment_df['plasmodium_species'] = plasm_assignment_df['plasmodium_species'].fillna('')
        plasm_spp = plasm_assignment_df.set_index('sample_id')['plasmodium_species'].to_dict()
        assert set(plasm_spp.keys()) == set(comb_stats_df.sample_id), \
            'plasmodium assignment samples do not match nn samples'

        # named species in legend - remove genus name
        plasm_legend_colors = {sp[11:]:color for sp, color in plasm_colors.iloc[:6].to_dict().items()}
        # other species collapsed
        assert plasm_colors.iloc[6:].nunique() == 1, \
            'plasmodium species color scheme not matching nn plot expectation'
        plasm_legend_colors['other'] = plasm_colors['unknown']
        # mixed/uninfected inferred during plotting
        plasm_legend_colors['mixed'] = '#000000'
        plasm_legend_colors['none'] = '#808080'

    # plot
    plates = comb_stats_df.plate_id.unique()
    nplates = comb_stats_df.plate_id.nunique()
    fig, axs = plt.subplots(nplates, 1, figsize=(20, 4 * nplates + 1))
    if nplates == 1:
        axs = [axs]
    for plate, ax in zip(plates, axs):
        plot_df = comb_stats_df[comb_stats_df.plate_id == plate].copy().sort_values('well_id').reset_index()
        plot_samples = plot_df['sample_id'] # TODO does not work for duplicated sample IDs
        # plot nn proportions for plate
        nn_level_result_plot_df = nn_level_result_df.loc[plot_samples]
        # TODO remove all-zero columns to speed up plotting - color index->label
        nn_level_result_plot_df.plot(
                kind='bar', stacked=True, width=1, ax=ax, color=level_colors
                )
        # remove zero size patches post-hoc
        for patch in ax.patches[:]:  # Copy list of patches to avoid modifying while iterating
            if isinstance(patch, mpatches.Rectangle):
                if patch.get_width() == 0 or patch.get_height() == 0:
                    patch.remove()
        ax.axhline(args.nn_assignment_threshold, color='k', ls=':', linewidth=1)
        ax.set_ylim(0, 1)
        ax.set_yticks([])
        ax.set_ylabel(plate, fontsize=16)
        handles, labels = ax.get_legend_handles_labels()
        ax.get_legend().remove()
        
        # bottom ticks - species name coloured by plasm
        ax.set_xticks(range(plot_df.shape[0]))
        ax.set_xticklabels(plot_df['sample_name'])
        if plasm_assignment_df is not None and plasm_colors is not None:
            for i, r in plot_df.iterrows():
                sample_plasm_sp = plasm_spp[r.sample_id]
                # multiple species infection
                if len(sample_plasm_sp.split(';')) > 1:
                    ax.get_xticklabels()[i].set_color('black')
                # species in index
                elif sample_plasm_sp in plasm_colors.keys():
                    ax.get_xticklabels()[i].set_color(plasm_colors[sample_plasm_sp])
                # no infection
                else:
                    ax.get_xticklabels()[i].set_color('grey')
        ax.tick_params(axis='x', rotation=90)
        
        # top ticks - targets coloured by multiallelics band
        ax2 = ax.twiny()
        ax2.set_xticks(range(plot_df.shape[0]))
        ax2.set_xticklabels(plot_df['mosq_targets_plot'])
        for i, r in plot_df.iterrows():
            ax2.get_xticklabels()[i].set_color(ma_colors[r.ma_band])
        ax2.tick_params(axis='x', rotation=90)
        ax2.set_xlim(ax.get_xlim())
    plt.tight_layout()

    # add legends after adjusting layout so that they span multiple subplots
    ma_artist = axs[0].legend(
        handles=[mpatches.Patch(color=color, label=label) for label, color in ma_colors.items()],
        title=f'Total target count (top number) colored\nby multiallelic targets (* locov, <{args.locov_rc} reads)',
        alignment='left',
        loc='upper left',
        bbox_to_anchor=(1,1.1),
        fontsize=10,
        ncols=2
    )
    axs[0].add_artist(ma_artist)

    if plasm_assignment_df is not None and plasm_colors is not None:
        plasm_artist = axs[0].legend(
            handles=[mpatches.Patch(color=color, label=label) for label, color in plasm_legend_colors.items()],
            title='Specimen ID label (bottom) colored by\nPlasmodium species detected',
            alignment='left',
            loc='upper left',
            bbox_to_anchor=(1,0.725),
            fontsize=10,
            ncols=2
        )
        axs[0].add_artist(plasm_artist)

    # subset legend to values observed over the cutoff
    if args.legend_cutoff > 0:
        logging.info(f'subsetting legend to observed labels at min frequency {args.legend_cutoff}')
    flt_handles = []
    flt_labels = []
    for handle, label in zip(handles, labels):
        max_freq = nn_level_result_df[label].max()
        if max_freq >= args.legend_cutoff:
            flt_handles.append(handle)
            flt_labels.append(label)
    # consistent legend title
    leg_title_labels = {
        'coarse':'Coarse',
        'int':'Intermediate',
        'fine':'Fine'
    }
    leg_title = f'{leg_title_labels[level_label]} NN level mosquito species assignments'
    if args.legend_cutoff > 0:
        leg_title += f'\nwith observed proportion over {args.legend_cutoff}'
    # reverse species legend order to match barplot order
    axs[0].legend(
        flt_handles[::-1], flt_labels[::-1], 
        title=leg_title,
        alignment='left',
        loc='upper left',
        bbox_to_anchor=(1,0.1), 
        fontsize=10
        )
    # adding title in post - handling margins by savefig's bbox_inches='tight' at this point
    axs[0].set_title(f'NN assignment {level_label} level for {run_id}', fontsize=20)

    return fig, axs


def nn(args):

    setup_logging(verbose=args.verbose)

    os.makedirs(args.outdir, exist_ok =True)

    logging.info('ANOSPP NN data import started')

    hap_df = prep_hap(args.haplotypes)
    run_id, comb_stats_df = prep_comb_stats(args.stats)
    
    logging.info(f'starting NN assignment for {comb_stats_df.sample_id.nunique()} samples in run {run_id}')
    mosq_hap_df = prep_mosquito_haps(
        hap_df,
        args.hap_read_count_threshold,
        args.hap_reads_fraction_threshold
        )

    ref_hap_df, allele_freqs, true_multi_targets, colors, version_name = prep_reference_index(
        args.reference_path
        )
        
    nn_hap_fn = f'{args.outdir}/nn_hap_summary.tsv'
    nndict_fn = f'{args.outdir}/nn_dist_to_ref.tsv'
    if args.resume and os.path.isfile(nndict_fn):
        logging.warning(f'reading nndict from {nndict_fn}')
        nndict = {}
        with open(nndict_fn) as f:
            next(f)
            for line in f:
                ll = line.strip().split('\t')
                if len(ll) == 3:
                    nndict[ll[0]] = ([int(i) for i in ll[1].split('|')], float(ll[2]))
        
    if args.resume and os.path.isfile(nn_hap_fn):
        logging.warning(f'reading annotated haplotype data from {nn_hap_fn}')
        non_error_hap_df = pd.read_csv(nn_hap_fn, sep='\t')
    else:
        kmers = construct_unique_kmer_table(mosq_hap_df, args.kmer_length, source='samples')
        ref_kmers = construct_unique_kmer_table(ref_hap_df, args.kmer_length, source=version_name)

        error_seqs = identify_error_seqs(mosq_hap_df, kmers, args.kmer_length, args.n_error_snps)
        non_error_hap_df = mosq_hap_df[~mosq_hap_df.seqid.isin(error_seqs)]
        non_error_hap_df = recompute_haplotype_coverage(non_error_hap_df)
        
        if not args.resume or not os.path.isfile(nndict_fn):
            nndict = find_nn_unique_haps(non_error_hap_df, kmers, ref_hap_df, ref_kmers)
            nn_df = pd.DataFrame.from_dict(nndict, orient='index', columns=['nn_id_array', 'nn_dist'])
            nn_df['nn_id'] = ['|'.join(map(str, l)) for l in nn_df.nn_id_array]
            nn_df.index.name = 'seqid'
            nn_df[['nn_id', 'nn_dist']].to_csv(
                nndict_fn, 
                sep='\t',
                index=True
                )
        nn_hap_df = add_nn_to_haplotypes(non_error_hap_df, nndict, ref_hap_df)
        logging.info(f'writing annotated haplotypes to {nn_hap_fn}')
        nn_hap_df.to_csv(nn_hap_fn, index=False, sep='\t')

    nn_assignment_fn = f'{args.outdir}/nn_assignment.tsv'
    if args.resume and os.path.isfile(nn_assignment_fn):
        logging.warning(f'reading nn assignments from {nn_assignment_fn}')
        nn_stats_df = pd.read_csv(nn_assignment_fn, sep='\t')
        comb_stats_df = pd.merge(comb_stats_df, nn_stats_df, on='sample_id', how='left')
        results_dfs = {}
        for level in ['coarse', 'int', 'fine']:
            level_assignment_fn = f'{args.outdir}/assignment_{level}.tsv'
            logging.warning(f'reading {level} assignments from {level_assignment_fn}')
            # TODO fix heterogeneity in use of sample_id as index
            results_dfs[level] = pd.read_csv(level_assignment_fn, sep='\t', index_col=0)
    else:
        results_dfs, test_samples = perform_nn_assignment_samples(
            non_error_hap_df, 
            ref_hap_df, 
            nndict, 
            allele_freqs, 
            args.allelism_normalisation
        )
        for level in ['coarse', 'int', 'fine']:
            results_dfs[level].to_csv(f'{args.outdir}/assignment_{level}.tsv', sep='\t')

        comb_stats_df = generate_hard_calls(
            comb_stats_df, 
            non_error_hap_df, 
            test_samples,
            results_dfs, 
            true_multi_targets, 
            nn_asgn_threshold=args.nn_assignment_threshold
        )

        comb_stats_df['nn_ref'] = version_name
        logging.info(f'writing assignment results to {nn_assignment_fn}')
        comb_stats_df[[
            'sample_id',
            'run_id',
            'multiallelic_mosq_targets',
            'mosq_reads',
            'mosq_targets_recovered',
            'nn_assignment',
            'nn_coarse',
            'nn_int',
            'nn_fine',
            'nn_species_call',
            'nn_call_method',
            'nn_ref'
        ]].to_csv(nn_assignment_fn, index=False, sep='\t')
    
    summary_text = generate_summary(comb_stats_df, version_name)
    summary_fn = f'{args.outdir}/nn_summary.txt'
    logging.info(f'writing summary file to {summary_fn}')
    with open(summary_fn, 'w') as fn:
        fn.write(summary_text)

    if not args.no_plotting:

        fig, _, _, _ = plot_het_cov(non_error_hap_df, title='Mosquito filtered', run_id=run_id)
        fig.savefig(f'{args.outdir}/het_cov.png')

        if args.plasm_assignment is not None and args.plasm_colors is not None:
            plasm_df = pd.read_csv(args.plasm_assignment, sep='\t')
            plasm_colors = pd.read_csv(args.plasm_colors).set_index('species')['color']
        else:
            plasm_df = None
            plasm_colors = None
        
        for level in ['coarse', 'int', 'fine']:
            fig_fn = f'{args.outdir}/{level}_assignment.png'
            if args.resume and os.path.isfile(fig_fn):
                logging.warning(f'nn figure {fig_fn} exists, not re-genrating')
            else:
                fig, _ = plot_assignment_proportions(
                    comb_stats_df, 
                    results_dfs[level], 
                    level, 
                    colors[level],
                    run_id,
                    plasm_df,
                    plasm_colors,
                    args
                    )
                fig.savefig(fig_fn, bbox_inches='tight')

    logging.info('ANOSPP NN complete')

    
def main():
    
    parser = argparse.ArgumentParser('NN assignment for ANOSPP sequencing data')
    parser.add_argument('-a', '--haplotypes', help='Haplotypes tsv file generatd by prep', required=True)
    parser.add_argument('-s', '--stats', help='Comb stats tsv file generated by prep', required=True)
    parser.add_argument('-r', '--reference_path', 
                        help='Path to reference index directory, expected to end with e.g. nnv2 ',
                        required=True)
    parser.add_argument('-o', '--outdir', help='Output directory. Default: nn', default='nn')
    parser.add_argument('--no_plotting', help='Do not generate plots. Default: False',
                        action='store_true',
                        default=False)
    parser.add_argument('--allelism_normalisation',
                        help='Normalisation method over multiple alleles. Options: [n_alleles,reads_fraction]. '
                        'Default: n_alleles',
                        choices=['n_alleles', 'reads_fraction'],
                        default='n_alleles')
    parser.add_argument('--hap_read_count_threshold',
                        help='Minimum number of reads for supported haplotypes.  Default: 10',
                        default=10, type=int)
    parser.add_argument('--hap_reads_fraction_threshold',
                        help='Minimum fraction of reads for supported haplotypes. Default: 0.1',
                        default=0.1, type=float)
    parser.add_argument('--locov_rc',
                        help='Samples with fewer than this number of reads are marked as low coverage in plot. Default: 1000',
                        default=1000, type=int)
    parser.add_argument('--nn_assignment_threshold',
                        help='Required fraction for calling assignment. Default: 0.7 (changed in nnv2)',
                        default=0.7, type=float)
    parser.add_argument('--n_error_snps',
                        help='Maximum number of snps for a multi-allelic sequence to be considered a sequencing or PCR error. Default: 2',
                        default=2, type=int)
    parser.add_argument('-k', '--kmer_length',
                        help='Length of k-mers to use. Note that NNoVAE has been developed and tested for k=8, '
                        'so accuracy of results cannot be guaranteed with other values of k. Default: k=8',
                        default=8, type=int)
    parser.add_argument('--plasm_assignment',
                        help='Path to plasm_assignment.tsv file used for sample label colouring '
                        'in nn plots. Default: None - colouring not applied',
                        default=None)
    parser.add_argument('--plasm_colors',
                        help='Path to species_colours.csv from plasm reference directory '
                        'used for sample label colouring in nn plots. '
                        'Default: None - colouring not applied',
                        default=None)
    parser.add_argument('--legend_cutoff',
                        help='Minimum observed NN assignment proportion '
                        'for species label to be added to the legend '
                        'Default: 0 - include all species in reference index',
                        default=0, type=float)
    parser.add_argument('--resume',
                        help='Do not re-generate nn_dist_to_ref.tsv and nn_assignment.tsv '
                        'if those are present in the output directory',
                        action='store_true', default=False)
    parser.add_argument('-v', '--verbose',
                        help='Include INFO level log messages',
                        action='store_true')

    args = parser.parse_args()
    args.outdir=args.outdir.rstrip('/')

    nn(args)


if __name__ == '__main__':
    main()


import pandas as pd
import logging

MOSQ_TARGETS = [str(i) for i in range(62)]
PLASM_TARGETS = ['P1', 'P2']
ANOSPP_TARGETS = MOSQ_TARGETS + PLASM_TARGETS
CUTADAPT_TARGETS = ANOSPP_TARGETS + ['unknown']

def setup_logging(verbose=False):
    try: 
        del logging.root.handlers[:]
    except:
        pass
    if verbose:
        logging.basicConfig(level=logging.INFO, format='[%(levelname)s] [%(asctime)s] %(message)s')
    else:
        logging.basicConfig(level=logging.WARNING, format='[%(levelname)s] [%(asctime)s] %(message)s')

def well_id_mapper():
    '''
    Yields mapping of tag_index 1,2...96 
    to well id A1,B1...H12.
    N.B. subsequent plates can be mapped using
    tag_index % 96, thus last well is inserted at 0
    '''

    well_ids = dict()
    tag = 1
    for col in range(1,13):
        for row in 'ABCDEFGH':
            well_ids[tag] = f'{row}{col}'
            tag += 1

    # edge case
    well_ids[0] = 'H12'
            
    return well_ids

def lims_well_id_mapper():
    '''
    Yields mapping of tag_index 1,2...384 
    to well id A1,C1...P24
    4 96-well plates order in quadrants is
    1 2
    3 4
    N.B. subsequent plates can be mapped using
    tag_index % 384, thus last well is inserted at 0
    '''
    lims_well_ids = dict()
    tag = 1
    # upper left quadrant
    for col in range(1,13):
        for row in 'ACEGIKMO':
            lims_well_ids[tag] = f'{row}{col * 2 - 1}'
            tag += 1
    # upper right quadrant
    for col in range(1,13):
        for row in 'ACEGIKMO':
            lims_well_ids[tag] = f'{row}{col * 2}'
            tag += 1
    # lower left quadrant
    for col in range(1,13):
        for row in 'BDFHJLNP':
            lims_well_ids[tag] = f'{row}{col * 2 - 1}'
            tag += 1
    # lower right quadrant
    for col in range(1,13):
        for row in 'BDFHJLNP':
            lims_well_ids[tag] = f'{row}{col * 2}'
            tag += 1

    # edge case
    lims_well_ids[0] = 'P24'

    return lims_well_ids

def seqid_generator(hap_df):
    '''
    assign identifiers to unique haplotypes
    '''

    seqids = dict()
    for tgt, group in hap_df.groupby('target'):
        for (i, cons) in enumerate(group['consensus'].unique()):
            seqids[tgt + cons] = '{}-{}'.format(tgt, i)
    hap_df['seqid'] = (hap_df.target + hap_df.consensus).replace(seqids)

    return hap_df

def human_format(num):
    if num is None:
        return ''
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0
    # 1 though 999
    if magnitude == 0:
        return '%.0f' % (num)
    # 1.0k through 99.9k
    elif num < 100:
        return '%.1f%s' % (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude])
    # 100k through 999k
    else:
        return '%.0f%s' % (num, ['', 'K', 'M', 'G', 'T', 'P'][magnitude])

def well_ordering(well_series):

    # well order for plotting - A1, B1, ...
    well_sequence = []
    for col in range(1,13):
        for row in 'ABCDEFGH': 
            well_sequence.append(f'{row}{col}')

    return pd.Categorical(well_series, categories=well_sequence)

def prep_hap(hap_fn, anospp=True):
    '''
    load haplotypes table generated by prep
    '''

    logging.info(f'loading haplotypes table from {hap_fn}')

    hap_df = pd.read_csv(hap_fn, sep='\t', dtype={'target':'str'})

    # compatibility with old style haplotype column names
    hap_df.rename(columns=({
        's_Sample':'sample_id',
        'frac_reads':'reads_fraction'
        }), 
    inplace=True)

    for col in ('sample_id',
                'target',
                'consensus',
                'reads'):
        assert col in hap_df.columns, f'hap column {col} not found'

    # for legacy - same as in anospp_analysis.prep.get_hap_df
    if 'total_reads' not in hap_df.columns:
        hap_df['total_reads'] = hap_df.groupby(by=['sample_id', 'target']) \
            ['reads'].transform('sum')

    if 'reads_fraction' not in hap_df.columns:
        hap_df['reads_fraction'] = hap_df['reads'] / hap_df['total_reads']

    if 'seqid' not in hap_df.columns:
        hap_df = seqid_generator(hap_df)

    if 'nalleles' not in hap_df.columns:
        hap_df['nalleles'] = hap_df.groupby(by=['sample_id', 'target']) \
            ['seqid'].transform('nunique')

    hap_df['consensus'] = hap_df['consensus'].str.upper()
    
    if hap_df['target'].isin(CUTADAPT_TARGETS).all():
        hap_df['target'] = pd.Categorical(hap_df['target'], 
                                        categories=CUTADAPT_TARGETS, 
                                        ordered=True)
    else:
        logging.warning('non-ANOSPP targets detected in haps, targets order might be unstable')
    
    hap_df.sort_values(by=[
        'sample_id',
        'target'
        ], inplace=True)

    return hap_df

def prep_comb_stats(comb_stats_fn, anospp=True):

    logging.info(f'loading combined stats table from {comb_stats_fn}')

    comb_stats_df = pd.read_csv(comb_stats_fn, sep='\t')

    assert comb_stats_df['run_id'].nunique(dropna=False) == 1, f'multiple runs in {comb_stats_fn}'

    run_id = comb_stats_df['run_id'].iloc[0]

    return run_id, comb_stats_df


import pandas as pd
import argparse

from anospp_analysis.util import *

def validate_aggregation(comb_df):

    logging.info('checking data types')

    for col in [
        'sample_id', #'irods_path', 'id_library_lims', 'id_study_lims', 'sanger_sample_id', 'front_barcode', 'rear_barcode'
        'run_id', 'lane_index', 'tag_index', 'plate_id',
        'well_id', 'lims_plate_id', 'lims_well_id', 'sample_name',
        'total_reads', 'readthrough_pass_reads', 'dada2_input_reads',
        'dada2_filtered_reads', 'dada2_denoised_reads', 'dada2_merged_reads',
        'dada2_nonchim_reads', 'target_reads', 'overall_filter_rate',
        'unassigned_asvs', 'targets_recovered', 'raw_mosq_targets_recovered',
        'raw_multiallelic_mosq_targets', 'raw_mosq_reads', 'p1_reads', 'p2_reads', 
        'p1_reads_pass', 'p2_reads_pass', 'plasmodium_detection_status', 'plasm_ref',
        # skip plasm hap
        'multiallelic_mosq_targets', 'mosq_reads', 'mosq_targets_recovered',
        'nn_assignment', 'nn_species_call', 'nn_call_method', 'nn_ref',
        # skip vae
        'nnovae_mosquito_species', 'nnovae_call_method'
        ]:
        assert ~comb_df[col].isna().any(), f'missing {col} values found'

    for col in [
        'run_id', 'lane_index', 'tag_index',
        'total_reads', 'readthrough_pass_reads', 'dada2_input_reads',
        'dada2_filtered_reads', 'dada2_denoised_reads', 'dada2_merged_reads',
        'dada2_nonchim_reads', 'target_reads',
        'unassigned_asvs', 'targets_recovered', 'raw_mosq_targets_recovered',
        'raw_multiallelic_mosq_targets', 'raw_mosq_reads', 'p1_reads', 'p2_reads', 
        'p1_reads_pass', 'p2_reads_pass',
        'multiallelic_mosq_targets', 'mosq_reads', 'mosq_targets_recovered'
        ]:
        assert pd.api.types.is_integer_dtype(comb_df[col]), f'{col} datatype is not integer'
        assert (comb_df[col] >= 0).all(), f'{col} contains negative values'

    for col in ['overall_filter_rate']:
        assert pd.api.types.is_numeric_dtype(comb_df[col]), f'{col} datatype is not numeric'

    # for col in ['mean1', 'mean2', 'mean3', 'sd1', 'sd2', 'sd3']:
    #     assert pd.api.types.is_numeric_dtype(comb_df[col]) or comb_df[col].empty, f'{col} datatype is not numeric'

    logging.info('checking columns contents')

    assert comb_df.sample_id.is_unique, 'duplicated sample_id found'

    assert len(comb_df.run_id.unique()) == 1, 'more than a single run_id found'

    assert comb_df.well_id.isin(well_id_mapper().values()).all(), 'non A1...H12 well_id found'

    assert comb_df.lims_well_id.isin(lims_well_id_mapper().values()).all(), 'non A1...P24 lims_well_id found'

    for (colp, coln) in [
        ('total_reads', 'readthrough_pass_reads'),
        ('readthrough_pass_reads', 'dada2_input_reads'),
        ('dada2_input_reads', 'dada2_filtered_reads'),
        ('dada2_filtered_reads', 'dada2_denoised_reads'),
        ('dada2_denoised_reads', 'dada2_merged_reads'),
        ('dada2_merged_reads', 'dada2_nonchim_reads'),
        ('dada2_nonchim_reads', 'target_reads'),
        ('p1_reads', 'p1_reads_pass'),
        ('p2_reads', 'p2_reads_pass'),
        ('raw_mosq_reads', 'mosq_reads'), # issue
        ('raw_mosq_targets_recovered', 'mosq_targets_recovered'), # issue
        ('raw_multiallelic_mosq_targets', 'multiallelic_mosq_targets')
        ]:
        assert (comb_df[colp] >= comb_df[coln]).all(), f'found less reads in {colp} than in {coln}'

    assert ((comb_df.overall_filter_rate >= 0) & (comb_df.overall_filter_rate <= 1)).all(), \
        'found overall_filter_rate outside of [0,1]'

    assert (comb_df.targets_recovered <= 64).all(), 'over 64 targets_recovered reported'

    for col in [
        'raw_mosq_targets_recovered', 'raw_multiallelic_mosq_targets',
        'multiallelic_mosq_targets', 'mosq_targets_recovered'
        ]:
        assert (comb_df[col] <= 62).all(), f'over 62 {col} reported'

    assert (comb_df.target_reads == comb_df.raw_mosq_reads + comb_df.p1_reads + comb_df.p2_reads).all(), \
        'target_reads does not match raw_mosq_reads + p1_reads + p2_reads'

    assert (comb_df.query('p1_reads_pass > 10')['p1_hapids_pass'].notna()).all(), \
        'not all P1 pass records supported by 10 reads have p1_hapids_pass recorded'
    assert ~((comb_df.p1_reads_pass < 10) & (comb_df.p1_reads_pass > 0)).any(), \
        'some P1 pass records supported by less than 10 reads'
    assert (comb_df.query('p1_reads_pass == 0')['p1_hapids_pass'].isna()).all(), \
        'record with zero p1_reads_pass has some p1_hapids_pass recorded'

    assert (comb_df.query('p2_reads_pass > 10')['p2_hapids_pass'].notna()).all(), \
        'not all P2 pass records supported by 10 reads have p2_hapids_pass recorded'
    assert ~((comb_df.p2_reads_pass < 10) & (comb_df.p2_reads_pass > 0)).any(), \
        'some P2 pass records supported by less than 10 reads'
    assert (comb_df.query('p2_reads_pass == 0')['p2_hapids_pass'].isna()).all(), \
        'record with zero p2_reads_pass has some p2_hapids_pass recorded'

def agg(args):

    setup_logging(verbose=args.verbose)

    logging.info('ANOSPP results merging data import started')
    
    run_id, comb_stats_df = prep_comb_stats(args.stats)
    # qc_df = pd.read_csv(args.qc, sep='\t')
    plasm_df = pd.read_csv(args.plasm, sep='\t')
    nn_df = pd.read_csv(args.nn, sep='\t')
    vae_df = pd.read_csv(args.vae, sep='\t')

    logging.info("merging results tables")

    # assert set(manifest_df.sample_id) == set(qc_df.sample_id), \
    #     'lanelets manifest and QC samples do not match'
    # comb_df = pd.merge(manifest_df, qc_df, how='inner')

    assert set(comb_stats_df.sample_id) == set(plasm_df.sample_id), \
        'plasm samples do not match comb stats'
    comb_df = pd.merge(comb_stats_df, plasm_df, how='inner')

    assert set(comb_df.sample_id) == set(nn_df.sample_id), \
        'NN samples do not match plasm, comb stats'
    comb_df = pd.merge(comb_df, nn_df, how='inner')

    assert vae_df.index.isin(comb_df.index).all(), \
        'VAE samples do not match NN, plasm, comb stats'
    comb_df = pd.merge(comb_df, vae_df, how='left')

    comb_df['nnovae_mosquito_species'] = comb_df.vae_species_call.fillna(comb_df.nn_species_call)
    is_nocall = comb_df['nnovae_mosquito_species'].isna()
    assert ~is_nocall.any(), \
        f'could not find none of NN or VAE call for {comb_df[is_nocall].index.to_list()}'
    
    # infer call method
    comb_df['nnovae_call_method'] = comb_df.nn_call_method
    comb_df.loc[
        comb_df.sample_id.isin(vae_df.sample_id),
        'nnovae_call_method'
    ] = 'VAE'

    if not args.force:
        validate_aggregation(comb_df)

    logging.info(f'writing merged results to {args.out}')
    comb_df.to_csv(args.out, sep='\t', index=False)


def main():
    
    parser = argparse.ArgumentParser('Merging ANOSPP run analysis results into a single file')
    parser.add_argument('-s', '--stats',
                        help='path to comb stats tsv generated by prep. Default: prep/comb_stats.tsv',
                        default='prep/comb_stats.tsv')
    parser.add_argument('-n', '--nn', 
                        help='path to NN assignment tsv. Default: nn/nn_assignment.tsv', 
                        default='nn/nn_assignment.tsv')
    parser.add_argument('-e', '--vae', 
                        help='path to VAE assignment tsv. Default: vae/vae_assignment.tsv',
                        default='vae/vae_assignment.tsv')
    parser.add_argument('-p', '--plasm', 
                        help='path to plasm assignment tsv. Default: plasm/plasm_assignment.tsv',
                        default='plasm/plasm_assignment.tsv')
    parser.add_argument('-o', '--out', 
                        help='Output aggregated sample metadata tsv. Default: anospp_results.tsv', 
                        default='anospp_results.tsv')
    parser.add_argument('-f', '--force', 
                        help='Skip aggregation validation', action='store_true', default=False)
    parser.add_argument('-v', '--verbose', 
                        help='Include INFO level log messages', action='store_true', default=True)
                        

    args = parser.parse_args()

    agg(args)


if __name__ == '__main__':
    main()